"""
Spectral analysis
=================

Perform a full region based on-off spectral analysis and fit the resulting datasets.

Prerequisites
-------------

-  Understanding how spectral extraction is performed in Cherenkov
   astronomy, in particular regarding OFF background measurements.
-  Understanding the basics data reduction and modeling/fitting process
   with the gammapy library API as shown in the :doc:`/tutorials/starting/analysis_2`
   tutorial.

Context
-------

While 3D analyses allow in principle to consider complex field of views
containing overlapping gamma-ray sources, in many cases we might have an
observation with a single, strong, point-like source in the field of
view. A spectral analysis, in that case, might consider all the events
inside a source (or ON) region and bin them in energy only, obtaining 1D
datasets.

In classical Cherenkov astronomy, the background estimation technique
associated with this method measures the number of events in OFF regions
taken in regions of the field-of-view devoid of gamma-ray emitters,
where the background rate is assumed to be equal to the one in the ON
region.

This allows to use a specific fit statistics for ON-OFF measurements,
the wstat (see `~gammapy.stats.wstat`), where no background model is
assumed. Background is treated as a set of nuisance parameters. This
removes some systematic effects connected to the choice or the quality
of the background model. But this comes at the expense of larger
statistical uncertainties on the fitted model parameters.

**Objective: perform a full region based spectral analysis of 4 Crab
observations of H.E.S.S. data release 1 and fit the resulting
datasets.**

Introduction
------------

Here, as usual, we use the `~gammapy.data.DataStore` to retrieve a
list of selected observations (`~gammapy.data.Observations`). Then, we
define the ON region containing the source and the geometry of the
`~gammapy.datasets.SpectrumDataset` object we want to produce. We then
create the corresponding dataset Maker.

We have to define the Maker object that will extract the OFF counts from
reflected regions in the field-of-view. To ensure we use data in an
energy range where the quality of the IRFs is good enough we also create
a safe range Maker.

We can then proceed with data reduction with a loop over all selected
observations to produce datasets in the relevant geometry.

We can then explore the resulting datasets and look at the cumulative
signal and significance of our source. We finally proceed with model
fitting.

In practice, we have to:

- Create a `~gammapy.data.DataStore` pointing to the relevant data
- Apply an observation selection to produce a list of observations,
  a `~gammapy.data.Observations` object.
- Define a geometry of the spectrum we want to produce:

  - Create a `~regions.CircleSkyRegion` for the ON extraction region
  - Create a `~gammapy.maps.MapAxis` for the energy binnings: one for the
    reconstructed (i.e.measured) energy, the other for the true energy
    (i.e.the one used by IRFs and models)

- Create the necessary makers :

  - the spectrum dataset maker : `~gammapy.makers.SpectrumDatasetMaker`
  - the OFF background maker, here a `~gammapy.makers.ReflectedRegionsBackgroundMaker`
  - and the safe range maker : `~gammapy.makers.SafeMaskMaker`

- Perform the data reduction loop. And for every observation:

  - Apply the makers sequentially to produce a `~gammapy.datasets.SpectrumDatasetOnOff`
  - Append it to list of datasets

- Define the `~gammapy.modeling.models.SkyModel` to apply to the dataset.
- Create a `~gammapy.modeling.Fit` object and run it to fit the model parameters
- Apply a `~gammapy.estimators.FluxPointsEstimator` to compute flux points for
  the spectral part of the fit.

"""

from pathlib import Path
import astropy.units as u
from astropy.coordinates import Angle, SkyCoord
from regions import CircleSkyRegion

# %matplotlib inline
import matplotlib.pyplot as plt

######################################################################
# Setup
# -----
#
# As usual, we’ll start with some setup …
#
from IPython.display import display
from gammapy.data import DataStore
from gammapy.datasets import (
    Datasets,
    FluxPointsDataset,
    SpectrumDataset,
)
from gammapy.estimators import FluxPointsEstimator
from gammapy.estimators.utils import resample_energy_edges
from gammapy.makers import (
    ReflectedRegionsBackgroundMaker,
    SafeMaskMaker,
    SpectrumDatasetMaker,
)
from gammapy.maps import MapAxis, RegionGeom, WcsGeom
from gammapy.modeling import Fit
from gammapy.modeling.models import (
    ExpCutoffPowerLawSpectralModel,
    SkyModel,
    create_crab_spectral_model,
)
from gammapy.visualization import plot_spectrum_datasets_off_regions


######################################################################
# Load Data
# ---------
#
# First, we select and load some H.E.S.S. observations of the Crab nebula.
#
# We will access the events, effective area, energy dispersion, livetime
# and PSF for containment correction.
#

datastore = DataStore.from_dir("$GAMMAPY_DATA/hess-dl3-dr1/")
obs_ids = [23523, 23526, 23559, 23592]
observations = datastore.get_observations(obs_ids)


######################################################################
# Define Target Region
# --------------------
#
# The next step is to define a signal extraction region, also known as on
# region. In the simplest case this is just a
# `~regions.CircleSkyRegion`.
#

target_position = SkyCoord(ra=83.63, dec=22.01, unit="deg", frame="icrs")
on_region_radius = Angle("0.11 deg")
on_region = CircleSkyRegion(center=target_position, radius=on_region_radius)


######################################################################
# Create exclusion mask
# ---------------------
#
# We will use the reflected regions method to place off regions to
# estimate the background level in the on region. To make sure the off
# regions don’t contain gamma-ray emission, we create an exclusion mask.
#
# Using http://gamma-sky.net/ we find that there’s only one known
# gamma-ray source near the Crab nebula: the AGN called `RGB
# J0521+212 <http://gamma-sky.net/#/cat/tev/23>`__ at GLON = 183.604 deg
# and GLAT = -8.708 deg.
#

exclusion_region = CircleSkyRegion(
    center=SkyCoord(183.604, -8.708, unit="deg", frame="galactic"),
    radius=0.5 * u.deg,
)

skydir = target_position.galactic
geom = WcsGeom.create(
    npix=(150, 150), binsz=0.05, skydir=skydir, proj="TAN", frame="icrs"
)

exclusion_mask = ~geom.region_mask([exclusion_region])
exclusion_mask.plot()
plt.show()


######################################################################
# Run data reduction chain
# ------------------------
#
# We begin with the configuration of the maker classes:
#

energy_axis = MapAxis.from_energy_bounds(
    0.1, 40, nbin=10, per_decade=True, unit="TeV", name="energy"
)
energy_axis_true = MapAxis.from_energy_bounds(
    0.05, 100, nbin=20, per_decade=True, unit="TeV", name="energy_true"
)

geom = RegionGeom.create(region=on_region, axes=[energy_axis])
dataset_empty = SpectrumDataset.create(geom=geom, energy_axis_true=energy_axis_true)

dataset_maker = SpectrumDatasetMaker(
    containment_correction=True, selection=["counts", "exposure", "edisp"]
)
bkg_maker = ReflectedRegionsBackgroundMaker(exclusion_mask=exclusion_mask)
safe_mask_maker = SafeMaskMaker(methods=["aeff-max"], aeff_percent=10)

datasets = Datasets()

for obs_id, observation in zip(obs_ids, observations):
    dataset = dataset_maker.run(dataset_empty.copy(name=str(obs_id)), observation)
    dataset_on_off = bkg_maker.run(dataset, observation)
    dataset_on_off = safe_mask_maker.run(dataset_on_off, observation)
    datasets.append(dataset_on_off)

print(datasets)

######################################################################
# The data reduction loop can also be performed through the
# `~gammapy.makers.DatasetsMaker` class that take a list of makers as input,
# as described :doc:`here </tutorials/details/makers>`


######################################################################
# Plot off regions
# ----------------
#

plt.figure()
ax = exclusion_mask.plot()
on_region.to_pixel(ax.wcs).plot(ax=ax, edgecolor="k")
plot_spectrum_datasets_off_regions(ax=ax, datasets=datasets)
plt.show()


######################################################################
# Source statistic
# ----------------
#
# Next we’re going to look at the overall source statistics in our signal
# region.
#

info_table = datasets.info_table(cumulative=True)

display(info_table)

######################################################################
# And make the corresponding plots

fig, (ax_excess, ax_sqrt_ts) = plt.subplots(figsize=(10, 4), ncols=2, nrows=1)
ax_excess.plot(
    info_table["livetime"].to("h"),
    info_table["excess"],
    marker="o",
    ls="none",
)

ax_excess.set_title("Excess")
ax_excess.set_xlabel("Livetime [h]")
ax_excess.set_ylabel("Excess events")

ax_sqrt_ts.plot(
    info_table["livetime"].to("h"),
    info_table["sqrt_ts"],
    marker="o",
    ls="none",
)

ax_sqrt_ts.set_title("Sqrt(TS)")
ax_sqrt_ts.set_xlabel("Livetime [h]")
ax_sqrt_ts.set_ylabel("Sqrt(TS)")
plt.show()


######################################################################
# Finally you can write the extracted datasets to disk using the OGIP (default)
# format (PHA, ARF, RMF, BKG, see
# `here <https://gamma-astro-data-formats.readthedocs.io/en/latest/spectra/ogip/index.html>`__
# for details):
#

path = Path("spectrum_analysis")
path.mkdir(exist_ok=True)
datasets.write(filename=path / "spectrum_dataset.yaml", overwrite=True)


######################################################################
# If you want to read back the datasets from disk you can use:
#

datasets = Datasets.read(filename=path / "spectrum_dataset.yaml")


######################################################################
# Fit spectrum
# ------------
#
# Now we’ll fit a global model to the spectrum. First we do a joint
# likelihood fit to all observations. If you want to stack the
# observations see below. We will also produce a debug plot in order to
# show how the global fit matches one of the individual observations.
#

spectral_model = ExpCutoffPowerLawSpectralModel(
    amplitude=1e-12 * u.Unit("cm-2 s-1 TeV-1"),
    index=2,
    lambda_=0.1 * u.Unit("TeV-1"),
    reference=1 * u.TeV,
)
model = SkyModel(spectral_model=spectral_model, name="crab")

datasets.models = [model]

fit_joint = Fit()
result_joint = fit_joint.run(datasets=datasets)


######################################################################
# Make a copy here to compare it later
#

model_best_joint = model.copy()


######################################################################
# Fit quality and model residuals
# -------------------------------
#


######################################################################
# We can access the results dictionary to see if the fit converged:
#

print(result_joint)


######################################################################
# and check the best-fit parameters
#

display(result_joint.models.to_parameters_table())


######################################################################
# A simple way to inspect the model residuals is using the function
# `~gammapy.datasets.SpectrumDataset.plot_fit()`
#

ax_spectrum, ax_residuals = datasets[0].plot_fit()
ax_spectrum.set_ylim(0.1, 40)
plt.show()


######################################################################
# For more ways of assessing fit quality, please refer to the dedicated
# :doc:`/tutorials/details/fitting` tutorial.
#


######################################################################
# Compute Flux Points
# -------------------
#
# To round up our analysis we can compute flux points by fitting the norm
# of the global model in energy bands.
# We create an instance of the
# `~gammapy.estimators.FluxPointsEstimator`, by passing the dataset and
# the energy binning:
#

fpe = FluxPointsEstimator(
    energy_edges=energy_axis.edges, source="crab", selection_optional="all"
)
flux_points = fpe.run(datasets=datasets)


######################################################################
# Here is a the table of the resulting flux points:
#

display(flux_points.to_table(sed_type="dnde", formatted=True))


######################################################################
# Now we plot the flux points and their likelihood profiles. For the
# plotting of upper limits we choose a threshold of TS < 4.
#

fig, ax = plt.subplots()
flux_points.plot(ax=ax, sed_type="e2dnde", color="darkorange")
flux_points.plot_ts_profiles(ax=ax, sed_type="e2dnde")
ax.set_xlim(0.6, 40)
plt.show()

######################################################################
# Note: it is also possible to plot the flux distribution with the spectral model overlaid,
# but you must ensure the axis binning is identical for the flux points and
# integral flux.
#


######################################################################
# The final plot with the best fit model, flux points and residuals can be
# quickly made like this:
#

flux_points_dataset = FluxPointsDataset(
    data=flux_points, models=model_best_joint.copy()
)
ax, _ = flux_points_dataset.plot_fit()
ax.set_xlim(0.6, 40)
plt.show()


######################################################################
# Stack observations
# ------------------
#
# An alternative approach to fitting the spectrum is stacking all
# observations first and the fitting a model. For this we first stack the
# individual datasets:
#

dataset_stacked = Datasets(datasets).stack_reduce()


######################################################################
# Again we set the model on the dataset we would like to fit (in this case
# it’s only a single one) and pass it to the `~gammapy.modeling.Fit`
# object:
#

dataset_stacked.models = model
stacked_fit = Fit()
result_stacked = stacked_fit.run([dataset_stacked])

######################################################################
# Make a copy to compare later
#

model_best_stacked = model.copy()

print(result_stacked)

######################################################################
# And display the parameter table for both the joint and the stacked models

display(model_best_joint.parameters.to_table())

display(model_best_stacked.parameters.to_table())


######################################################################
# Finally, we compare the results of our stacked analysis to a previously
# published Crab Nebula Spectrum for reference. This is available in
# `~gammapy.modeling.models.create_crab_spectral_model`.
#
fig, ax = plt.subplots()

plot_kwargs = {
    "energy_bounds": [0.1, 30] * u.TeV,
    "sed_type": "e2dnde",
    "ax": ax,
}
ax.yaxis.set_units(u.Unit("erg cm-2 s-1"))
# plot stacked model
model_best_stacked.spectral_model.plot(**plot_kwargs, label="Stacked analysis result")
model_best_stacked.spectral_model.plot_error(facecolor="blue", alpha=0.3, **plot_kwargs)

# plot joint model
model_best_joint.spectral_model.plot(
    **plot_kwargs, label="Joint analysis result", ls="--"
)
model_best_joint.spectral_model.plot_error(facecolor="orange", alpha=0.3, **plot_kwargs)

create_crab_spectral_model("hess_ecpl").plot(
    **plot_kwargs,
    label="Crab reference",
)
ax.legend()
plt.show()

# sphinx_gallery_thumbnail_number = 5

######################################################################
# A note on statistics
# --------------------
#
# Different statistic are available for the `~gammapy.datasets.FluxPointsDataset` :
#
# - chi2 : estimate from chi2 statistics.
# - profile : estimate from interpolation of the likelihood profile.
# - distrib : estimate from probability distributions, assuming that flux points
#   correspond to asymmetric gaussians and upper limits complementary error functions.
#
# Default is `chi2`, in that case upper limits are ignored and the mean of asymmetric error is used.
# So it is recommended to use `profile` if `stat_scan` is available on flux points.
# The `distrib` case provides an approximation if the `profile` is not available
# which allows to take into accounts upper limit and asymmetric error.
#
# In the example below we can see that the `profile` case matches exactly the result
# from the joint analysis of the ON/OFF datasets using `wstat` (as labelled).


def plot_stat(fp_dataset):
    fig, ax = plt.subplots()

    plot_kwargs = {
        "energy_bounds": [0.1, 30] * u.TeV,
        "sed_type": "e2dnde",
        "ax": ax,
    }

    fp_dataset.data.plot(energy_power=2, ax=ax)
    model_best_joint.spectral_model.plot(
        color="b", lw=0.5, **plot_kwargs, label="wstat"
    )

    stat_types = ["chi2", "profile", "distrib"]
    colors = ["red", "g", "c"]
    lss = ["--", ":", "--"]

    for ks, stat in enumerate(stat_types):
        fp_dataset.stat_type = stat

        fit = Fit()
        fit.run([fp_dataset])

        fp_dataset.models[0].spectral_model.plot(
            color=colors[ks], ls=lss[ks], **plot_kwargs, label=stat
        )
        fp_dataset.models[0].spectral_model.plot_error(
            facecolor=colors[ks], **plot_kwargs
        )
        plt.legend()


plot_stat(flux_points_dataset)

######################################################################
#
# In order to avoid discrepancies due to the treatment of upper limits
# we can utilise the `~gammapy.estimators.utils.resample_energy_edges`
# for defining energy bins in which the minimum number of `sqrt_ts` is 2.
# In that case all the statistics definitions give equivalent results.
#

energy_edges = resample_energy_edges(dataset_stacked, conditions={"sqrt_ts_min": 2})

fpe_no_ul = FluxPointsEstimator(
    energy_edges=energy_edges, source="crab", selection_optional="all"
)
flux_points_no_ul = fpe_no_ul.run(datasets=datasets)
flux_points_dataset_no_ul = FluxPointsDataset(
    data=flux_points_no_ul,
    models=model_best_joint.copy(),
)

plot_stat(flux_points_dataset_no_ul)

######################################################################
# What next?
# ----------
#
# The methods shown in this tutorial is valid for point-like or slightly
# extended sources where we can assume that the IRF taken at the region
# center is valid over the whole region. If one wants to extract the 1D
# spectrum of a large source and properly average the response over the
# extraction region, one has to use a different approach explained in
# the :doc:`/tutorials/analysis-1d/extended_source_spectral_analysis`
# tutorial.
#



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Utility script to work with the codemeta.json file."""
import json
import logging
from configparser import ConfigParser
import click
from datetime import date

log = logging.getLogger(__name__)


def update_codemeta(maintainer, filename, setup_file=None):
    log.info(f"Updating {filename}")
    # add potentially missing content
    with open(filename, "r") as f:
        data = json.load(f)

    for author in data["author"]:
        if "familyName" in author.keys() and author["familyName"] == maintainer:
            log.info(f"Setting maintainer to {maintainer}")
            data["maintainer"] = author

    data["readme"] = "https://gammapy.org"
    data["issueTracker"] = "https://github.com/gammapy/gammapy/issues"
    data["developmentStatus"] = ("active",)
    data["email"] = "GAMMAPY-COORDINATION-L@IN2P3.FR"

    modified_date = str(date.today())
    data["dateModified"] = modified_date


    if setup_file:
        # complete with software requirements from setup.cfg
        cf = ConfigParser()
        cf.read(setup_file)
        requirements = cf["options"]["install_requires"].split("\n")
        if "" in requirements:
            requirements.remove("")
        data["softwareRequirements"] = requirements

    with open(filename, "w") as f:
        json.dump(data, f, indent=4)

    # replace bad labelled attributes
    with open(filename, "r") as f:
        content = f.read()

    content = content.replace("legalName", "name")
    content = content.replace("version", "softwareVersion")

    with open(filename, "w") as f:
        f.write(content)


@click.command()
@click.option("--maintainer", default="Donath", type=str, help="Maintainer name")
@click.option(
    "--filename", default="../codemeta.json", type=str, help="codemeta filename"
)
@click.option("--setup_file", default="../setup.cfg", type=str, help="setup filename")
@click.option(
    "--log_level",
    default="INFO",
    type=click.Choice(["DEBUG", "INFO", "WARNING"]),
    help="log level",
)
def cli(maintainer, filename, log_level, setup_file):
    logging.basicConfig(level=log_level)
    log.setLevel(level=log_level)
    update_codemeta(maintainer=maintainer, filename=filename, setup_file=setup_file)


if __name__ == "__main__":
    cli()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import logging as log
import numpy as np
from scipy.interpolate import CubicSpline
from scipy.optimize import curve_fit
from scipy.stats import norm
from astropy.visualization import make_lupton_rgb
import matplotlib.axes as maxes
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable

__all__ = [
    "add_colorbar",
    "plot_contour_line",
    "plot_map_rgb",
    "plot_theta_squared_table",
    "plot_distribution",
]


ARTIST_TO_LINE_PROPERTIES = {
    "color": "markeredgecolor",
    "edgecolor": "markeredgecolor",
    "ec": "markeredgecolor",
    "facecolor": "markerfacecolor",
    "fc": "markerfacecolor",
    "linewidth": "markeredgewidth",
    "lw": "markeredgewidth",
}


def add_colorbar(img, ax, axes_loc=None, **kwargs):
    """
    Add colorbar to a given axis.

    Parameters
    ----------
    img : `~matplotlib.image.AxesImage`
        The image to plot the colorbar for.
    ax : `~matplotlib.axes.Axes`
        Matplotlib axes.
    axes_loc : dict, optional
        Keyword arguments passed to `~mpl_toolkits.axes_grid1.axes_divider.AxesDivider.append_axes`.
    kwargs : dict, optional
        Keyword arguments passed to `~matplotlib.pyplot.colorbar`.

    Returns
    -------
    cbar : `~matplotlib.pyplot.colorbar`
        The colorbar.

    Examples
    --------
    .. testcode::

        from gammapy.maps import Map
        from gammapy.visualization import add_colorbar
        import matplotlib.pyplot as plt
        map_ = Map.read("$GAMMAPY_DATA/cta-1dc-gc/cta-1dc-gc.fits.gz")
        axes_loc = {"position": "right", "size": "2%", "pad": "10%"}
        kwargs_colorbar = {'label':'Colorbar label'}

        # Example outside gammapy
        fig = plt.figure(figsize=(6, 3))
        ax = fig.add_subplot(111)
        img = ax.imshow(map_.sum_over_axes().data[0,:,:])
        add_colorbar(img, ax=ax, axes_loc=axes_loc, **kwargs_colorbar)

        # `add_colorbar` is available for the `plot` function here:
        fig = plt.figure(figsize=(6, 3))
        ax = fig.add_subplot(111)
        map_.sum_over_axes().plot(ax=ax, add_cbar=True, axes_loc=axes_loc,
                                  kwargs_colorbar=kwargs_colorbar)  # doctest: +SKIP

    """
    kwargs.setdefault("use_gridspec", True)
    kwargs.setdefault("orientation", "vertical")

    axes_loc = axes_loc or {}
    axes_loc.setdefault("position", "right")
    axes_loc.setdefault("size", "5%")
    axes_loc.setdefault("pad", "2%")
    axes_loc.setdefault("axes_class", maxes.Axes)

    divider = make_axes_locatable(ax)
    cax = divider.append_axes(**axes_loc)
    cbar = plt.colorbar(img, cax=cax, **kwargs)
    return cbar


def plot_map_rgb(map_, ax=None, **kwargs):
    """
    Plot RGB image on matplotlib WCS axes.

    This function is based on the `~astropy.visualization.make_lupton_rgb` function. The input map must
    contain 1 non-spatial axis with exactly 3 bins. If this is not the case, the map has to be resampled
    before using the `plot_map_rgb` function (e.g. as shown in the code example below).

    Parameters
    ----------
    map_ : `~gammapy.maps.WcsNDMap`
        WCS map. The map must contain 1 non-spatial axis with exactly 3 bins.
    ax : `~astropy.visualization.wcsaxes.WCSAxes`, optional
        WCS axis object to plot on.
    **kwargs : dict
        Keyword arguments passed to `~astropy.visualization.make_lupton_rgb`.

    Returns
    -------
    ax : `~astropy.visualization.wcsaxes.WCSAxes`
        WCS axis object.

    Examples
    --------
    >>> from gammapy.visualization import plot_map_rgb
    >>> from gammapy.maps import Map, MapAxis
    >>> import astropy.units as u
    >>> map_ = Map.read("$GAMMAPY_DATA/cta-1dc-gc/cta-1dc-gc.fits.gz")
    >>> axis_rgb = MapAxis.from_energy_edges(
    ...     [0.1, 0.2, 0.5, 10], unit=u.TeV, name="energy", interp="log"
    ... )
    >>> map_ = map_.resample_axis(axis_rgb)
    >>> kwargs = {"stretch": 0.5, "Q": 1, "minimum": 0.15}
    >>> plot_map_rgb(map_.smooth(0.08*u.deg), **kwargs) #doctest: +SKIP
    """
    geom = map_.geom
    if len(geom.axes) != 1 or geom.axes[0].nbin != 3:
        raise ValueError(
            "One non-spatial axis with exactly 3 bins is needed to plot an RGB image"
        )

    data = [data_slice / np.nanmax(data_slice.flatten()) for data_slice in map_.data]
    data = make_lupton_rgb(*data, **kwargs)

    ax = map_._plot_default_axes(ax=ax)
    ax.imshow(data)

    if geom.is_allsky:
        ax = map_._plot_format_allsky(ax)
    else:
        ax = map_._plot_format(ax)

    # without this the axis limits are changed when calling scatter
    ax.autoscale(enable=False)

    return ax


def plot_contour_line(ax, x, y, **kwargs):
    """Plot smooth curve from contour points."""
    xf = x
    yf = y

    # close contour
    if not (x[0] == x[-1] and y[0] == y[-1]):
        xf = np.append(x, x[0])
        yf = np.append(y, y[0])

    # curve parametrization must be strictly increasing
    # so we use the cumulative distance of each point from the first one
    dist = np.sqrt(np.diff(xf) ** 2.0 + np.diff(yf) ** 2.0)
    dist = [0] + list(dist)
    t = np.cumsum(dist)
    ts = np.linspace(0, t[-1], 50)

    # 1D cubic spline interpolation
    cs = CubicSpline(t, np.c_[xf, yf], bc_type="periodic")
    out = cs(ts)

    # plot
    if "marker" in kwargs.keys():
        marker = kwargs.pop("marker")
    else:
        marker = "+"
    if "color" in kwargs.keys():
        color = kwargs.pop("color")
    else:
        color = "b"

    ax.plot(out[:, 0], out[:, 1], "-", color=color, **kwargs)
    ax.plot(xf, yf, linestyle="", marker=marker, color=color)


def plot_theta_squared_table(table):
    """Plot the theta2 distribution of counts, excess and significance.

    Take the table containing the ON counts, the OFF counts, the acceptance,
    the off acceptance and the alpha (normalisation between ON and OFF)
    for each theta2 bin.

    Parameters
    ----------
    table : `~astropy.table.Table`
        Required columns: theta2_min, theta2_max, counts, counts_off and alpha
    """
    from gammapy.maps import MapAxis
    from gammapy.maps.axes import UNIT_STRING_FORMAT
    from gammapy.maps.utils import edges_from_lo_hi

    theta2_edges = edges_from_lo_hi(
        table["theta2_min"].quantity, table["theta2_max"].quantity
    )
    theta2_axis = MapAxis.from_edges(theta2_edges, interp="lin", name="theta_squared")

    ax0 = plt.subplot(2, 1, 1)

    x = theta2_axis.center.value
    x_edges = theta2_axis.edges.value
    xerr = (x - x_edges[:-1], x_edges[1:] - x)

    ax0.errorbar(
        x,
        table["counts"],
        xerr=xerr,
        yerr=np.sqrt(table["counts"]),
        linestyle="None",
        label="Counts",
    )

    ax0.errorbar(
        x,
        table["counts_off"],
        xerr=xerr,
        yerr=np.sqrt(table["counts_off"]),
        linestyle="None",
        label="Counts Off",
    )

    ax0.errorbar(
        x,
        table["excess"],
        xerr=xerr,
        yerr=(table["excess_errn"], table["excess_errp"]),
        fmt="+",
        linestyle="None",
        label="Excess",
    )

    ax0.set_ylabel("Counts")
    ax0.set_xticks([])
    ax0.set_xlabel("")
    ax0.legend()

    ax1 = plt.subplot(2, 1, 2)
    ax1.errorbar(x, table["sqrt_ts"], xerr=xerr, linestyle="None")
    ax1.set_xlabel(f"Theta [{theta2_axis.unit.to_string(UNIT_STRING_FORMAT)}]")
    ax1.set_ylabel("Significance")


def plot_distribution(
    wcs_map,
    mask=None,
    ax=None,
    ncols=3,
    func=None,
    kwargs_hist=None,
    kwargs_axes=None,
    kwargs_fit=None,
):
    """
    Plot the 1D distribution of data inside a map as an histogram. If the dimension of the map is smaller than 2,
    a unique plot will be displayed. Otherwise, if the dimension is 3 or greater, a grid of plot will be displayed.

    Parameters
    ----------
    wcs_map : `~gammapy.maps.WcsNDMap`
        A map that contains data to be plotted.
    mask : `~gammapy.maps.WcsNDMap`, optional
        2D mask defining the input data region.
    ax : `~matplotlib.axes.Axes` or list of `~matplotlib.axes.Axes`
        Axis object to plot on. If a list of Axis is provided it has to be the same length as the length of _map.data.
    ncols : int
        Number of columns to plot if a "plot grid" was to be done.
    func : function object or str
        The function used to fit a map data histogram or "norm". Default is None.
        If None, no fit will be performed. If "norm" is given, `scipy.stats.norm.pdf`
        will be passed to `scipy.optimize.curve_fit`.
    kwargs_hist : dict
        Keyword arguments to pass to `matplotlib.pyplot.hist`.
    kwargs_axes : dict
        Keyword arguments to pass to `matplotlib.axes.Axes`.
    kwargs_fit : dict
        Keyword arguments to pass to `scipy.optimize.curve_fit`

    Returns
    -------
    axes : `~numpy.ndarray` of `~matplotlib.pyplot.Axes`
        Array of Axes.
    result_list : list of dict
        List of dictionnary that contains the results of `scipy.optimize.curve_fit`. The number of elements in the list
        correspond to the dimension of the non-spatial axis of the map.
        The dictionnary contains:

            * `axis_edges` : the edges of the non-spatial axis bin used
            * `param` : the best-fit parameters of the input function `func`
            * `covar` : the covariance matrix for the fitted parameters `param`
            * `info_dict` : the `infodict` return of `scipy.optimize.curve_fit`

    Examples
    --------
    >>> from gammapy.datasets import MapDataset
    >>> from gammapy.estimators import TSMapEstimator
    >>> from scipy.stats import norm
    >>> from gammapy.visualization import plot_distribution
    >>> dataset = MapDataset.read("$GAMMAPY_DATA/cta-1dc-gc/cta-1dc-gc.fits.gz")
    >>> tsmap_est = TSMapEstimator().run(dataset)
    >>> axs, res = plot_distribution(tsmap_est.sqrt_ts, func="norm", kwargs_hist={'bins': 75, 'range': (-10, 10), 'density': True})
    >>> # Equivalently, one can do the following:
    >>> func = lambda x, mu, sig : norm.pdf(x, loc=mu, scale=sig)
    >>> axs, res = plot_distribution(tsmap_est.sqrt_ts, func=func, kwargs_hist={'bins': 75, 'range': (-10, 10), 'density': True})
    """
    from gammapy.maps import WcsNDMap  # import here to avoid circular import

    if not isinstance(wcs_map, WcsNDMap):
        raise TypeError(
            f"map_ must be an instance of gammapy.maps.WcsNDMap, given {type(wcs_map)}"
        )

    kwargs_hist = kwargs_hist or {}
    kwargs_axes = kwargs_axes or {}
    kwargs_fit = kwargs_fit or {}

    kwargs_hist.setdefault("density", True)
    kwargs_fit.setdefault("full_output", True)

    cutout, cutout_mask = wcs_map.cutout_and_mask_region()

    if mask:
        if mask.geom != wcs_map.geom.to_image():
            raise ValueError("Map and mask spatial geometry must agree!")

        if not mask.is_mask:
            raise ValueError(
                f"Mask map must be of boolean type, got {mask.data.dtype} instead."
            )

        cutout_mask.data = np.logical_and(cutout_mask.data, mask.data)

    idx_x, idx_y = np.where(cutout_mask)

    data = cutout.data[..., idx_x, idx_y]

    if ax is None:
        n_plot = len(data)
        cols = min(ncols, n_plot)
        rows = 1 + (n_plot - 1) // cols

        width = 12
        figsize = (width, width * rows / cols)

        fig, axes = plt.subplots(
            nrows=rows,
            ncols=cols,
            figsize=figsize,
        )
        cells_in_grid = rows * cols
    else:
        axes = np.array([ax])
        cells_in_grid = axes.size

    if not isinstance(axes, np.ndarray):
        axes = np.array([axes])

    result_list = []

    for idx in range(cells_in_grid):
        axe = axes.flat[idx]
        if idx > len(data) - 1:
            axe.set_visible(False)
            continue
        d = data[idx][np.isfinite(data[idx])]
        n, bins, _ = axe.hist(d, **kwargs_hist)

        if func is not None:
            kwargs_plot_fit = {"label": "Fit"}
            centers = 0.5 * (bins[1:] + bins[:-1])

            if func == "norm":

                def func_to_fit(x, mu, sigma):
                    return norm.pdf(x, mu, sigma)

                pars, cov, infodict, message, _ = curve_fit(
                    func_to_fit, centers, n, **kwargs_fit
                )

                mu, sig = pars[0], pars[1]
                err_mu, err_sig = np.sqrt(cov[0][0]), np.sqrt(cov[1][1])

                label_norm = (
                    r"$\mu$ = {:.2f} ± {:.2E}\n$\sigma$ = {:.2f} ± {:.2E}".format(
                        mu, err_mu, sig, err_sig
                    )
                ).replace(r"\n", "\n")
                kwargs_plot_fit["label"] = label_norm

            else:
                func_to_fit = func

                pars, cov, infodict, message, _ = curve_fit(
                    func_to_fit, centers, n, **kwargs_fit
                )

            axis_edges = (
                wcs_map.geom.axes[-1].edges[idx],
                wcs_map.geom.axes[-1].edges[idx + 1],
            )
            result_dict = {
                "axis_edges": axis_edges,
                "param": pars,
                "covar": cov,
                "info_dict": infodict,
            }
            result_list.append(result_dict)
            log.info(message)

            xmin, xmax = kwargs_hist.get("range", (np.min(d), np.max(d)))
            x = np.linspace(xmin, xmax, 1000)

            axe.plot(x, func_to_fit(x, *pars), lw=2, color="black", **kwargs_plot_fit)

        axe.set(**kwargs_axes)
        axe.legend()

    return axes, result_list



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import logging
import re
import numpy as np
from astropy.table import Table
from astropy.time import Time
import click
from github import Github, GithubException

log = logging.getLogger(__name__)


class GitHubInfoExtractor:
    """Class to interact with GitHub and extract PR and issues info tables.

    Parameters
    ----------
    repo : str
        input repository. Default is 'gammapy/gammapy'
    token : str
        GitHub access token. Default is None
    """

    def __init__(self, repo=None, token=None):
        self.repo = repo if repo else "gammapy/gammapy"
        self.github = self.login(token)
        self.repo = self.github.get_repo(repo)

    @staticmethod
    def login(token):
        if token:
            g = Github(token, per_page=100)
        else:
            g = Github()

        try:
            user_login = g.get_user().login
        except GithubException:
            user_login = "anonymous"

        log.info(f"Logging in GitHub as {user_login}")
        return g

    def check_requests_number(self):
        remaining, total = self.github.rate_limiting
        log.info(f"Remaining {remaining} requests over {total} requests.")

    @staticmethod
    def _get_commits_info(commits):
        """Builds a dictionary containing the number of commits and the list of unique committers."""
        result = dict()
        result["commits_number"] = commits.totalCount
        committers = set()
        for commit in commits:
            if commit.committer:
                committers.add(commit.committer.login)
        result["unique_committers"] = list(committers)
        return result

    @staticmethod
    def _get_reviews_info(reviews):
        """Builds a dictionary containing the number of reviews and the list of unique reviewers."""
        result = dict()
        result["review_number"] = reviews.totalCount
        reviewers = set()
        for review in reviews:
            if review.user:
                reviewers.add(review.user.login)
        result["unique_reviewers"] = list(reviewers)
        return result

    def _extract_pull_request_info(self, pull_request):
        """Builds a dictionary containing a list of summary informations.

        Parameters
        ----------
        pull_request :
            input pull request object

        Returns
        -------
        info : dict
            the result dictionary
        """
        result = dict()

        result["number"] = pull_request.number
        result["title"] = pull_request.title
        result["milestone"] = (
            "" if not pull_request.milestone else pull_request.milestone.title
        )
        result["is_merged"] = pull_request.is_merged()
        creation = pull_request.created_at
        result["date_creation"] = Time(creation) if creation else None
        closing = pull_request.closed_at
        result["date_closed"] = Time(closing) if closing else None
        result["user_name"] = pull_request.user.name
        result["user_login"] = pull_request.user.login
        result["user_email"] = pull_request.user.email
        result["labels"] = [label.name for label in pull_request.labels]
        result["changed_files"] = pull_request.changed_files
        result["base"] = pull_request.base.ref

        # extract commits
        commits = pull_request.get_commits()
        result.update(self._get_commits_info(commits))
        # extract reviews
        reviews = pull_request.get_reviews()
        result.update(self._get_reviews_info(reviews))

        return result

    def extract_pull_requests_table(
        self, state="closed", number_min=1, include_backports=False
    ):
        """Extract list of Pull Requests and build info table.

         Parameters
         ----------
        state : str ("closed", "open", "all")
             state of PRs to extract.
         number_min : int
             minimum PR number to include. Default is 0.
         include_backports : bool
             Include backport PRs in the table. Default is True.
        """
        pull_requests = self.repo.get_pulls(
            state=state, sort="created", direction="desc"
        )

        self.check_requests_number()

        results = []

        for pr in pull_requests:
            number = pr.number
            if number <= number_min:
                log.info(f"Reached minimum PR number {number_min}.")
                break

            title = pr.title
            if not include_backports and "Backport" in title:
                log.info(f"Pull Request {number} is backport. Skipping")
                continue

            log.info(f"Extracting Pull Request {number}.")
            try:
                result = self._extract_pull_request_info(pr)
            except AttributeError:
                log.warning(f"Issue with Pull Request {number}. Skipping")
                continue
            results.append(result)

        table = Table(results)
        return table
        self.check_requests_number()


@click.group()
@click.option(
    "--log-level",
    default="INFO",
    type=click.Choice(["DEBUG", "INFO", "WARNING"]),
)
def cli(log_level):
    logging.basicConfig(level=log_level)
    log.setLevel(level=log_level)


@cli.command("create_pull_request_table", help="Dump a table of all PRs.")
@click.option("--token", default=None, type=str)
@click.option("--repo", default="gammapy/gammapy", type=str)
@click.option("--state", default="closed", type=str)
@click.option("--number_min", default=4000, type=int)
@click.option("--filename", default="table_pr.ecsv", type=str)
@click.option("--overwrite", default=False, type=bool)
@click.option("--include_backports", default=False, type=bool)
def create_pull_request_table(
    repo, token, state, number_min, filename, overwrite, include_backports
):
    """Extract PR table and write it to dosk."""
    extractor = GitHubInfoExtractor(repo=repo, token=token)
    table = extractor.extract_pull_requests_table(
        state=state, number_min=number_min, include_backports=include_backports
    )
    table.write(filename, overwrite=overwrite)


@cli.command("merged_PR", help="Make a summary of PRs merged with a given milestone")
@click.argument("filename", type=str, default="table_pr.ecsv")
@click.argument("milestones", type=str, nargs=-1)
@click.option("--from_backports", default=False, type=bool)
def list_merged_PRs(filename, milestones, from_backports):
    """Make a list of merged PRs."""
    log.info(
        f"Make list of merged PRs from milestones {milestones} from file {filename}."
    )
    table = Table.read(filename)

    # Keep only merged PRs
    table = table[table["is_merged"] == True]

    # Keep the requested milestones
    valid = np.zeros((len(table)), dtype="bool")

    for i, pr in enumerate(table):
        milestone = milestones[0]
        if from_backports and "Backport" in pr["title"]:
            # check that the branch and milestone match
            if np.all(pr["base"].split(".")[:-1] == milestone.split(".")[:-1]):
                pattern = r"#(\d+)"
                parent_pr_number = int(re.search(pattern, pr["title"]).group(1))
                idx = np.where(table["number"] == parent_pr_number)[0]
                valid[idx] = True
        elif pr["milestone"] == milestone:
            valid[i] = True

    # filter the table and print info
    table = table[valid]
    log.info(f"Found {len(table)} merged PRs in the table.")

    unique_names = set()
    names = table["user_name"]
    logins = table["user_login"]
    for name, login in zip(names, logins):
        unique_names.add(name if name else login)

    unique_committers = set()
    unique_reviewers = set()
    for pr in table:
        for committer in pr["unique_committers"]:
            unique_committers.add(committer)
        for reviewer in pr["unique_reviewers"]:
            unique_reviewers.add(reviewer)

    contributor_names = list(unique_names)
    log.info(f"Found {len(contributor_names)} contributors in the table.")
    log.info(f"Found {len(unique_committers)} committers in the table.")
    log.info(f"namely: {unique_committers}")
    log.info(f"Found {len(unique_reviewers)} reviewers in the table.")
    log.info(f"namely: {unique_reviewers}")

    result = "Contributors\n"
    result += "~~~~~~~~~~~~\n"
    for name in contributor_names:
        result += f"- {name}\n"

    result += "\n\nPull Requests\n"
    result += "~~~~~~~~~~~~~\n\n"
    result += "This list is incomplete. Small improvements and bug fixes are not listed here.\n"

    for pr in table:
        number = pr["number"]
        title = pr["title"]
        user = pr["user_name"] if pr["user_name"] is not None else pr["user_login"]
        result += f"- [#{number}] {title} ({user})\n"

    print(result)


if __name__ == "__main__":
    cli()



"""Example how to write a command line tool with Click"""

import click
from gammapy.stats import CashCountsStatistic


# You can call the callback function for the click command anything you like.
# `cli` is just a commonly used generic term for "command line interface".
@click.command()
@click.argument("n_observed", type=float)
@click.argument("mu_background", type=float)
@click.option(
    "--value",
    type=click.Choice(["sqrt_ts", "p_value"]),
    default="sqrt_ts",
    help="Significance or p_value",
)
def cli(n_observed, mu_background, value):
    """Compute significance for a Poisson count observation.

    The significance is the tail probability to observe N_OBSERVED counts
    or more, given a known background level MU_BACKGROUND."""
    stat = CashCountsStatistic(n_observed, mu_background)
    if value == "sqrt_ts":
        s = stat.sqrt_ts
    else:
        s = stat.p_value

    print(s)


if __name__ == "__main__":
    cli()



r"""
.. _PhaseCurve-temporal-model:

Phase curve temporal model
==========================

This model parametrises a PhaseCurve time model, i.e. with a template phasogram and timing parameters

"""

import astropy.units as u
from astropy.time import Time
from gammapy.modeling.models import (
    Models,
    PowerLawSpectralModel,
    SkyModel,
    TemplatePhaseCurveTemporalModel,
)

path = "$GAMMAPY_DATA/tests/phasecurve_LSI_DC.fits"
t_ref = 43366.275 * u.d
f0 = 1.0 / (26.7 * u.d)

phase_model = TemplatePhaseCurveTemporalModel.read(path, t_ref, 0.0, f0)
time_range = [Time("59100", format="mjd"), Time("59200", format="mjd")]

phase_model.plot(time_range, n_points=400)

# %%
# YAML representation
# -------------------
# Here is an example YAML file using the model:

model = SkyModel(
    spectral_model=PowerLawSpectralModel(),
    temporal_model=phase_model,
    name="phase_curve_model",
)
models = Models([model])

print(models.to_yaml())



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import logging
import pytest
import numpy as np
from numpy.testing import assert_allclose
import astropy.units as u
from astropy.coordinates import EarthLocation, SkyCoord
from astropy.io import fits
from astropy.table import Table
from astropy.time import Time
from gammapy.data import GTI, DataStore, Observation
from gammapy.data.pointing import FixedPointingInfo
from gammapy.datasets import MapDataset, MapDatasetEventSampler
from gammapy.datasets.tests.test_map import get_map_dataset
from gammapy.irf import load_irf_dict_from_file
from gammapy.makers import MapDatasetMaker
from gammapy.maps import MapAxis, RegionGeom, RegionNDMap, WcsGeom
from gammapy.modeling.models import (
    ConstantSpectralModel,
    FoVBackgroundModel,
    GaussianSpatialModel,
    LightCurveTemplateTemporalModel,
    Models,
    PointSpatialModel,
    PowerLawSpectralModel,
    SkyModel,
)
from gammapy.utils.testing import requires_data

LOCATION = EarthLocation(lon="-70d18m58.84s", lat="-24d41m0.34s", height="2000m")


@pytest.fixture()
def signal_model():
    spatial_model = GaussianSpatialModel(
        lon_0="0 deg", lat_0="0 deg", sigma="0.2 deg", frame="galactic"
    )

    spectral_model = PowerLawSpectralModel(amplitude="1e-11 cm-2 s-1 TeV-1")

    t_max = 1000 * u.s

    time = np.arange(t_max.value) * u.s
    tau = u.Quantity("2e2 s")
    norm = np.exp(-time / tau)

    table = Table()
    table["TIME"] = time
    table["NORM"] = norm / norm.max()
    t_ref = Time("2000-01-01")
    table.meta = dict(MJDREFI=t_ref.mjd, MJDREFF=0, TIMEUNIT="s", TIMESYS="utc")
    temporal_model = LightCurveTemplateTemporalModel.from_table(table)

    return SkyModel(
        spatial_model=spatial_model,
        spectral_model=spectral_model,
        temporal_model=temporal_model,
        name="test-source",
    )


@pytest.fixture()
def models(signal_model):
    bkg_model = FoVBackgroundModel(dataset_name="test")
    return [signal_model, bkg_model]


@pytest.fixture()
def model_alternative():
    spatial_model1 = GaussianSpatialModel(
        lon_0="0 deg", lat_0="0 deg", sigma="0.2 deg", frame="galactic"
    )

    spectral_model = PowerLawSpectralModel(amplitude="1e-11 cm-2 s-1 TeV-1")

    mod1 = SkyModel(
        spatial_model=spatial_model1,
        spectral_model=spectral_model,
        name="test-source",
    )

    spatial_model2 = GaussianSpatialModel(
        lon_0="0.5 deg", lat_0="0.5 deg", sigma="0.2 deg", frame="galactic"
    )

    mod2 = SkyModel(
        spatial_model=spatial_model2,
        spectral_model=spectral_model,
        name="test-source2",
    )

    spatial_model3 = GaussianSpatialModel(
        lon_0="0.5 deg", lat_0="0.0 deg", sigma="0.2 deg", frame="galactic"
    )

    mod3 = SkyModel(
        spatial_model=spatial_model3,
        spectral_model=spectral_model,
        name="test-source3",
    )

    bkg_model = FoVBackgroundModel(dataset_name="test")

    model2 = Models([mod1, mod2, bkg_model, mod3])
    return model2


def get_energy_dependent_temporal_model():
    energy_axis = MapAxis.from_energy_bounds(
        energy_min=1 * u.TeV, energy_max=10 * u.TeV, nbin=10, name="energy"
    )

    edges = np.linspace(0, 1000, 101) * u.s
    time_axis = MapAxis.from_edges(edges=edges, name="time", interp="lin")

    geom = RegionGeom.create(
        region=SkyCoord("0.5 deg", "0.5 deg", frame="galactic"),
        axes=[energy_axis, time_axis],
    )

    coords = geom.get_coord(sparse=True)

    def f(time, energy):
        """Toy model for testing"""
        amplitude = u.Quantity("5e-11 cm-2 s-1 TeV-1") * np.exp(-(time / (200 * u.s)))
        index = 1 + 2.2 * time.to_value("s") / 1000
        return PowerLawSpectralModel.evaluate(
            amplitude=amplitude, index=index, energy=energy, reference=1 * u.TeV
        )

    data = f(time=coords["time"], energy=coords["energy"])

    m = RegionNDMap.from_geom(data=data.value, geom=geom, unit=data.unit)
    t_ref = Time(51544.00074287037, format="mjd", scale="tt")
    return LightCurveTemplateTemporalModel(m, t_ref=t_ref)


@pytest.fixture()
@requires_data()
def energy_dependent_temporal_sky_model(models):
    models[0].spatial_model = PointSpatialModel(
        lon_0="0 deg", lat_0="0 deg", frame="galactic"
    )
    models[0].spectral_model = ConstantSpectralModel(const="1 cm-2 s-1 TeV-1")
    models[0].temporal_model = get_energy_dependent_temporal_model()
    return models[0]


@pytest.fixture(scope="session")
def dataset():
    energy_axis = MapAxis.from_bounds(
        1, 10, nbin=3, unit="TeV", name="energy", interp="log"
    )

    geom = WcsGeom.create(
        skydir=(0, 0), binsz=0.05, width="5 deg", frame="galactic", axes=[energy_axis]
    )

    etrue_axis = energy_axis.copy(name="energy_true")
    geom_true = geom.to_image().to_cube(axes=[etrue_axis])

    dataset = get_map_dataset(
        geom=geom, geom_etrue=geom_true, edisp="edispmap", name="test"
    )
    dataset.background /= 400

    dataset.gti = GTI.create(
        start=0 * u.s, stop=1000 * u.s, reference_time=Time("2000-01-01").tt
    )

    return dataset


@requires_data()
def test_evaluate_timevar_source(energy_dependent_temporal_sky_model, dataset):
    dataset.models = energy_dependent_temporal_sky_model
    evaluator = dataset.evaluators["test-source"]

    sampler = MapDatasetEventSampler(random_state=0)
    npred = sampler._evaluate_timevar_source(dataset, evaluator.model)

    assert_allclose(np.shape(npred.data), (3, 1999, 1, 1))

    assert_allclose(npred.data[:, 10, 0, 0], [0.819546, 1.676989, 2.248684], rtol=2e-4)
    assert_allclose(npred.data[:, 50, 0, 0], [0.729098, 1.442345, 1.869792], rtol=2e-4)

    filename = "$GAMMAPY_DATA/gravitational_waves/GW_example_DC_map_file.fits.gz"
    temporal_model = LightCurveTemplateTemporalModel.read(filename, format="map")
    temporal_model.t_ref.value = 51544.00074287037
    dataset.models[0].temporal_model = temporal_model
    evaluator = dataset.evaluators["test-source"]

    sampler = MapDatasetEventSampler(random_state=0)
    npred = sampler._evaluate_timevar_source(dataset, evaluator.model)

    assert_allclose(
        npred.data[:, 1000, 0, 0] / 1e-13,
        [0.038113, 0.033879, 0.010938],
        rtol=2e-4,
    )


@requires_data()
def test_sample_coord_time_energy_no_spatial(
    dataset, energy_dependent_temporal_sky_model
):
    sampler = MapDatasetEventSampler(random_state=0)

    energy_dependent_temporal_sky_model.spatial_model = None
    energy_dependent_temporal_sky_model.spectral_model = ConstantSpectralModel(
        const="1 cm-2 s-1 TeV-1"
    )
    dataset.models = energy_dependent_temporal_sky_model
    evaluator = dataset.evaluators["test-source"]

    with pytest.raises(TypeError):
        sampler._sample_coord_time_energy(dataset, evaluator.model)


@requires_data()
def test_sample_coord_time_energy_gaussian(
    dataset, energy_dependent_temporal_sky_model
):
    sampler = MapDatasetEventSampler(random_state=0)

    energy_dependent_temporal_sky_model.spatial_model = GaussianSpatialModel()
    energy_dependent_temporal_sky_model.spectral_model = ConstantSpectralModel(
        const="1 cm-2 s-1 TeV-1"
    )
    dataset.models = energy_dependent_temporal_sky_model
    evaluator = dataset.evaluators["test-source"]

    with pytest.raises(TypeError):
        sampler._sample_coord_time_energy(dataset, evaluator.model)


@requires_data()
def test_sample_coord_time_energy(dataset, energy_dependent_temporal_sky_model):
    sampler = MapDatasetEventSampler(random_state=1)

    energy_dependent_temporal_sky_model.spatial_model = PointSpatialModel(
        lon_0="0 deg", lat_0="0 deg", frame="galactic"
    )
    energy_dependent_temporal_sky_model.spectral_model.const.value = 2

    dataset.models = energy_dependent_temporal_sky_model
    evaluator = dataset.evaluators["test-source"]

    time_ref = dataset.gti.time_ref

    expected = np.array([7.840697, 266.404988, -28.936178])
    events = sampler._sample_coord_time_energy(dataset, evaluator.model)
    met = (events["TIME"] - time_ref).to_value("s")

    assert len(events) == 2514
    assert_allclose(met[0], 854.26361, rtol=1e-6)
    assert_allclose((events[0][1], events[0][2], events[0][3]), expected, rtol=1e-6)

    irfs = load_irf_dict_from_file(
        "$GAMMAPY_DATA/cta-1dc/caldb/data/cta/1dc/bcf/South_z20_50h/irf_file.fits"
    )
    livetime = 1.0 * u.hr
    pointing = FixedPointingInfo(
        fixed_icrs=SkyCoord(0, 0, unit="deg", frame="galactic").icrs,
    )
    obs = Observation.create(
        obs_id=1001,
        pointing=pointing,
        livetime=livetime,
        irfs=irfs,
        location=LOCATION,
    )

    new_mod = Models(
        [
            energy_dependent_temporal_sky_model,
            FoVBackgroundModel(dataset_name=dataset.name),
        ]
    )
    dataset.models = new_mod
    events = sampler.run(dataset, obs)
    assert dataset.gti.time_ref.scale == events.table.meta["TIMESYS"]


@requires_data()
def test_fail_sample_coord_time_energy(
    dataset, models, energy_dependent_temporal_sky_model
):
    new_dataset = dataset.copy("my-dataset")
    new_dataset.gti = GTI.create(
        start=0 * u.s, stop=2.5 * u.s, reference_time=Time("2000-01-01").tt
    )

    new_dataset.models = energy_dependent_temporal_sky_model
    new_dataset.models[0].temporal_model.map.data *= 1e20
    evaluator = new_dataset.evaluators["test-source"]

    sampler = MapDatasetEventSampler(random_state=0, oversample_energy_factor=1)
    with pytest.raises(ValueError):
        sampler._sample_coord_time_energy(new_dataset, evaluator.model)


@requires_data()
def test_negative_npred(dataset):
    spatial_model = PointSpatialModel(lon_0="0 deg", lat_0="0 deg", frame="galactic")
    spectral_model = PowerLawSpectralModel(amplitude="-4e-10 cm-2 s-1 TeV-1")

    dataset.models = [
        SkyModel(spectral_model=spectral_model, spatial_model=spatial_model),
        FoVBackgroundModel(dataset_name=dataset.name),
    ]

    sampler = MapDatasetEventSampler(random_state=0, n_event_bunch=1000)
    events = sampler.run(dataset=dataset)

    assert len(events.table) == 15


@requires_data()
def test_sample_coord_time_energy_random_seed(
    dataset, energy_dependent_temporal_sky_model
):
    sampler = MapDatasetEventSampler(random_state=2)

    energy_dependent_temporal_sky_model.temporal_model.map._unit == ""
    dataset.models = energy_dependent_temporal_sky_model
    evaluator = dataset.evaluators["test-source"]

    events = sampler._sample_coord_time_energy(dataset, evaluator.model)

    assert len(events) == 1256

    assert_allclose(
        [events[0][1], events[0][2], events[0][3]],
        [1.932196, 266.404988, -28.936178],
        rtol=1e-3,
    )

    # Important: do not increase the tolerance!
    met = events["TIME"] - dataset.gti.time_ref
    assert_allclose(met[0].to_value("s"), 0.29982, rtol=1.5e-6)


@requires_data()
def test_sample_coord_time_energy_unit(dataset, energy_dependent_temporal_sky_model):
    sampler = MapDatasetEventSampler(random_state=1)

    energy_dependent_temporal_sky_model.temporal_model.map._unit == "cm-2 s-1 TeV-1"
    energy_dependent_temporal_sky_model.spectral_model.parameters[0].unit = ""
    dataset.models = energy_dependent_temporal_sky_model
    evaluator = dataset.evaluators["test-source"]

    events = sampler._sample_coord_time_energy(dataset, evaluator.model)

    assert len(events) == 1254
    assert_allclose(
        [events[0][1], events[0][2], events[0][3]],
        [6.22904, 266.404988, -28.936178],
        rtol=1e-6,
    )

    # Important: do not increase the tolerance!
    met = events["TIME"] - dataset.gti.time_ref
    assert_allclose(met[0].to_value("s"), 854.10859, rtol=1.5e-6)


@requires_data()
def test_mde_sample_sources(dataset, models):
    dataset.models = models
    sampler = MapDatasetEventSampler(random_state=0)
    events_table = sampler.sample_sources(dataset=dataset)

    assert len(events_table["ENERGY_TRUE"]) == 90
    assert_allclose(events_table["ENERGY_TRUE"][0], 2.383778805, rtol=1e-5)
    assert events_table["ENERGY_TRUE"].unit == "TeV"

    assert_allclose(events_table["RA_TRUE"][0], 266.56408893, rtol=1e-5)
    assert events_table["RA_TRUE"].unit == "deg"

    assert_allclose(events_table["DEC_TRUE"][0], -28.748145, rtol=1e-5)
    assert events_table["DEC_TRUE"].unit == "deg"

    assert_allclose(events_table["TIME"][0].mjd, 51544.0013961193, rtol=1e-9)

    assert_allclose(events_table["MC_ID"][0], 1, rtol=1e-5)


@requires_data()
def test_mde_sample_sources_psf_update(dataset, models):
    dataset.models = models
    sampler = MapDatasetEventSampler(random_state=0)
    events_table = sampler.sample_sources(dataset=dataset, psf_update=False)
    assert len(events_table["ENERGY_TRUE"]) == 90


@requires_data()
def test_sample_sources_energy_dependent(dataset, energy_dependent_temporal_sky_model):
    dataset.models = energy_dependent_temporal_sky_model

    sampler = MapDatasetEventSampler(random_state=0)
    events_table = sampler.sample_sources(dataset=dataset)

    assert len(events_table["ENERGY_TRUE"]) == 1268
    assert_allclose(events_table["ENERGY_TRUE"][0], 6.456526, rtol=1e-5)

    assert_allclose(events_table["RA_TRUE"][0], 266.404988, rtol=1e-5)

    assert_allclose(events_table["DEC_TRUE"][0], -28.936178, rtol=1e-5)

    assert_allclose(events_table["TIME"][0].mjd, 51544.001847785, rtol=1e-9)

    dt = np.max(events_table["TIME"]) - np.min(events_table["TIME"])
    assert (
        dt.to_value("s")
        <= dataset.gti.time_sum.to("s").value + sampler.t_delta.to("s").value
    )


@requires_data()
def test_mde_sample_weak_src(dataset, models):
    irfs = load_irf_dict_from_file(
        "$GAMMAPY_DATA/cta-1dc/caldb/data/cta/1dc/bcf/South_z20_50h/irf_file.fits"
    )
    livetime = 10.0 * u.hr
    pointing = FixedPointingInfo(
        fixed_icrs=SkyCoord(0, 0, unit="deg", frame="galactic").icrs,
    )
    obs = Observation.create(
        obs_id=1001,
        pointing=pointing,
        livetime=livetime,
        irfs=irfs,
        location=LOCATION,
    )

    models[0].parameters["amplitude"].value = 1e-25

    dataset.models = models

    sampler = MapDatasetEventSampler(random_state=0)
    events = sampler.run(dataset=dataset, observation=obs)

    assert len(events.table) == 18
    assert_allclose(
        len(np.where(events.table["MC_ID"] == 0)[0]), len(events.table), rtol=1e-5
    )


@requires_data()
def test_mde_sample_background(dataset, models):
    dataset.models = models
    sampler = MapDatasetEventSampler(random_state=0)
    events_table = sampler.sample_background(dataset=dataset)

    assert len(events_table) == 15

    assert_allclose(np.mean(events_table["ENERGY"]), 4.1, rtol=0.1)
    assert events_table["ENERGY"].unit == "TeV"

    assert np.all(events_table["RA"] < 269.95)
    assert np.all(events_table["RA"] > 262.90)
    assert events_table["RA"].unit == "deg"

    assert np.all(events_table["DEC"] < -25.43)
    assert np.all(events_table["DEC"] > -32.43)
    assert events_table["DEC"].unit == "deg"

    assert events_table["DEC_TRUE"][0] == events_table["DEC"][0]

    assert_allclose(events_table["MC_ID"][0], 0, rtol=1e-5)


@requires_data()
def test_mde_sample_psf(dataset, models):
    dataset.models = models
    sampler = MapDatasetEventSampler(random_state=0)
    events = sampler.sample_sources(dataset=dataset)
    events_table = sampler.sample_psf(dataset.psf, events)

    assert len(events_table) == 90
    assert_allclose(events_table["ENERGY_TRUE"][0], 2.38377880, rtol=1e-5)
    assert events_table["ENERGY_TRUE"].unit == "TeV"

    assert_allclose(events_table["RA"][0], 266.542912, rtol=1e-5)
    assert events_table["RA"].unit == "deg"

    assert_allclose(events_table["DEC"][0], -28.78829, rtol=1e-5)
    assert events_table["DEC"].unit == "deg"


@requires_data()
def test_mde_sample_edisp(dataset, models):
    dataset.models = models
    sampler = MapDatasetEventSampler(random_state=0)
    events = sampler.sample_sources(dataset=dataset)
    events_table = sampler.sample_edisp(dataset.edisp, events)

    assert len(events_table) == 90
    assert_allclose(events_table["ENERGY"][0], 2.383778805, rtol=1e-5)
    assert events_table["ENERGY"].unit == "TeV"

    assert_allclose(events_table["RA_TRUE"][0], 266.564088, rtol=1e-5)
    assert events_table["RA_TRUE"].unit == "deg"

    assert_allclose(events_table["DEC_TRUE"][0], -28.7481450, rtol=1e-5)
    assert events_table["DEC_TRUE"].unit == "deg"

    assert_allclose(events_table["MC_ID"][0], 1, rtol=1e-5)


@requires_data()
def test_event_det_coords(dataset, models):
    irfs = load_irf_dict_from_file(
        "$GAMMAPY_DATA/cta-1dc/caldb/data/cta/1dc/bcf/South_z20_50h/irf_file.fits"
    )
    livetime = 1.0 * u.hr
    pointing = FixedPointingInfo(
        fixed_icrs=SkyCoord(0, 0, unit="deg", frame="galactic").icrs,
    )
    obs = Observation.create(
        obs_id=1001,
        pointing=pointing,
        livetime=livetime,
        irfs=irfs,
        location=LOCATION,
    )

    dataset.models = models
    sampler = MapDatasetEventSampler(random_state=0)
    events = sampler.run(dataset=dataset, observation=obs)

    assert len(events.table) == 99

    # Check that det coordinates are within sqrt(2) * width of dataset
    assert np.all(events.table["DETX"] < 3.53)
    assert np.all(events.table["DETX"] > -3.53)

    assert events.table["DETX"].unit == "deg"

    assert np.all(events.table["DETY"] < 3.53)
    assert np.all(events.table["DETY"] > -3.53)

    assert events.table["DETY"].unit == "deg"


@requires_data()
def test_mde_run(dataset, models, caplog, tmp_path):
    irfs = load_irf_dict_from_file(
        "$GAMMAPY_DATA/cta-1dc/caldb/data/cta/1dc/bcf/South_z20_50h/irf_file.fits"
    )
    livetime = 1.0 * u.hr
    pointing = FixedPointingInfo(
        fixed_icrs=SkyCoord(0, 0, unit="deg", frame="galactic").icrs,
    )
    obs = Observation.create(
        obs_id=1001,
        pointing=pointing,
        livetime=livetime,
        irfs=irfs,
        location=LOCATION,
    )

    dataset.models = models

    logging.getLogger().setLevel(logging.INFO)
    sampler = MapDatasetEventSampler(random_state=0)
    events = sampler.run(dataset=dataset, observation=obs)

    captured = caplog.records
    str = [
        "Evaluating model: test-source",
        "Evaluating background...",
        "Event sampling completed.",
    ]

    assert captured[1].message == str[0]
    assert captured[2].message == str[1]
    assert captured[3].message == str[2]

    dataset_bkg = dataset.copy(name="new-dataset")
    dataset_bkg.models = [FoVBackgroundModel(dataset_name=dataset_bkg.name)]

    events_bkg = sampler.run(dataset=dataset_bkg, observation=obs)

    #    assert len(events.table) == 99
    assert_allclose(np.mean(events.table["ENERGY"]), 3.5, atol=1.0)
    assert np.all(events.table["ENERGY"] > 1)

    src_events = events.select_parameter("MC_ID", [0.5, 1.5])
    assert_allclose(np.mean(src_events.table["RA"]), 266.40, atol=0.1)
    assert_allclose(np.mean(src_events.table["DEC"]), -28.93, atol=0.1)
    separation = src_events.radec.separation(models[0].spatial_model.position)
    assert np.all(separation.to_value("deg") < 0.7)

    assert len(events_bkg.table) == 21
    assert_allclose(events_bkg.table["MC_ID"][0], 0, rtol=1e-5)

    meta = events.table.meta

    assert meta["HDUCLAS1"] == "EVENTS"
    assert meta["EXTNAME"] == "EVENTS"
    assert (
        meta["HDUDOC"]
        == "https://github.com/open-gamma-ray-astro/gamma-astro-data-formats"
    )
    assert meta["HDUVERS"] == "0.2"
    assert meta["HDUCLASS"] == "GADF"
    assert meta["OBS_ID"] == 1001
    assert_allclose(meta["TSTART"], 0.0)
    assert_allclose(meta["TSTOP"], 3600.0)
    assert_allclose(meta["ONTIME"], 3600.0)
    assert_allclose(meta["LIVETIME"], 3600.0)
    assert_allclose(meta["DEADC"], 1.0)
    assert_allclose(meta["RA_PNT"], 266.4049882865447)
    assert_allclose(meta["DEC_PNT"], -28.936177761791473)
    assert meta["EQUINOX"] == "J2000"
    assert meta["RADECSYS"] == "icrs"
    assert "Gammapy" in meta["CREATOR"]
    assert meta["EUNIT"] == "TeV"
    assert meta["EVTVER"] == ""
    assert meta["OBSERVER"] == "Gammapy user"
    assert meta["DSTYP1"] == "TIME"
    assert meta["DSUNI1"] == "s"
    assert meta["DSVAL1"] == "TABLE"
    assert meta["DSREF1"] == ":GTI"
    assert meta["DSTYP2"] == "ENERGY"
    assert meta["DSUNI2"] == "TeV"
    assert ":" in meta["DSVAL2"]
    assert meta["DSTYP3"] == "POS(RA,DEC)     "
    assert "CIRCLE" in meta["DSVAL3"]
    assert meta["DSUNI3"] == "deg             "
    assert meta["NDSKEYS"] == " 3 "
    assert_allclose(meta["RA_OBJ"], 266.4049882865447)
    assert_allclose(meta["DEC_OBJ"], -28.936177761791473)
    assert_allclose(meta["TELAPSE"], 1000.0)
    assert_allclose(meta["MJDREFI"], 51544)
    assert_allclose(meta["MJDREFF"], 0.0007428703684126958)
    assert meta["TIMEUNIT"] == "s"
    assert meta["TIMESYS"] == "tt"
    assert meta["TIMEREF"] == "LOCAL"
    assert meta["DATE-OBS"] == "2000-01-01"
    assert meta["DATE-END"] == "2000-01-01"
    assert meta["CONV_DEP"] == 0
    assert meta["CONV_RA"] == 0
    assert meta["CONV_DEC"] == 0
    assert meta["MID00000"] == 0
    assert meta["MMN00000"] == "test-bkg"
    assert meta["MID00001"] == 1
    assert meta["NMCIDS"] == 2
    assert_allclose(meta["ALT_PNT"], -13.5345076464, rtol=1e-7)
    assert_allclose(meta["AZ_PNT"], 228.82981620065763, rtol=1e-7)
    assert meta["ORIGIN"] == "Gammapy"
    assert meta["TELESCOP"] == "CTA"
    assert meta["INSTRUME"] == "1DC"
    assert meta["N_TELS"] == ""
    assert meta["TELLIST"] == ""

    # test writing out and reading back in works
    obs.events = events
    path = tmp_path / "obs.fits.gz"
    obs.write(path)
    obs_back = Observation.read(path)
    assert u.isclose(obs_back.observatory_earth_location.lon, LOCATION.lon)
    assert u.isclose(obs_back.observatory_earth_location.lat, LOCATION.lat)
    assert u.isclose(obs_back.observatory_earth_location.height, LOCATION.height)


@requires_data()
def test_irf_alpha_config(dataset, models):
    irfs = load_irf_dict_from_file(
        "$GAMMAPY_DATA/cta-caldb/Prod5-South-20deg-AverageAz-14MSTs37SSTs.180000s-v0.1.fits.gz"
    )
    livetime = 1.0 * u.hr
    pointing = FixedPointingInfo(
        fixed_icrs=SkyCoord(0, 0, unit="deg", frame="galactic").icrs,
    )
    obs = Observation.create(
        obs_id=1001,
        pointing=pointing,
        livetime=livetime,
        irfs=irfs,
        location=LOCATION,
    )

    dataset.models = models
    sampler = MapDatasetEventSampler(random_state=0)
    events = sampler.run(dataset=dataset, observation=obs)
    assert events is not None


@requires_data()
def test_mde_run_switchoff(dataset, models):
    irfs = load_irf_dict_from_file(
        "$GAMMAPY_DATA/cta-1dc/caldb/data/cta/1dc/bcf/South_z20_50h/irf_file.fits"
    )
    livetime = 1.0 * u.hr
    pointing = FixedPointingInfo(
        fixed_icrs=SkyCoord(0, 0, unit="deg", frame="galactic").icrs,
    )
    obs = Observation.create(
        obs_id=1001,
        pointing=pointing,
        livetime=livetime,
        irfs=irfs,
        location=LOCATION,
    )

    dataset.models = models

    dataset.psf = None
    dataset.edisp = None
    dataset.background = None

    sampler = MapDatasetEventSampler(random_state=0)
    events = sampler.run(dataset=dataset, observation=obs)

    assert len(events.table) == 90
    assert_allclose(events.table["ENERGY"][0], 1.947042, rtol=1e-5)
    assert_allclose(events.table["RA"][0], 266.875015, rtol=1e-5)
    assert_allclose(events.table["DEC"][0], -29.115063, rtol=1e-5)

    meta = events.table.meta

    assert meta["RA_PNT"] == 266.4049882865447
    assert_allclose(meta["ONTIME"], 3600.0)
    assert meta["OBS_ID"] == 1001
    assert meta["RADECSYS"] == "icrs"


@requires_data()
def test_events_datastore(tmp_path, dataset, models):
    irfs = load_irf_dict_from_file(
        "$GAMMAPY_DATA/cta-1dc/caldb/data/cta/1dc/bcf/South_z20_50h/irf_file.fits"
    )
    livetime = 10.0 * u.hr
    pointing = FixedPointingInfo(
        fixed_icrs=SkyCoord(0, 0, unit="deg", frame="galactic").icrs,
    )
    obs = Observation.create(
        obs_id=1001,
        pointing=pointing,
        livetime=livetime,
        irfs=irfs,
        location=LOCATION,
    )

    dataset.models = models
    sampler = MapDatasetEventSampler(random_state=0)
    events = sampler.run(dataset=dataset, observation=obs)

    primary_hdu = fits.PrimaryHDU()
    hdu_evt = fits.BinTableHDU(events.table)
    hdu_gti = dataset.gti.to_table_hdu(format="gadf")
    hdu_all = fits.HDUList([primary_hdu, hdu_evt, hdu_gti])
    hdu_all.writeto(str(tmp_path / "events.fits"))

    DataStore.from_events_files([str(tmp_path / "events.fits")])


@requires_data()
def test_MC_ID(model_alternative):
    irfs = load_irf_dict_from_file(
        "$GAMMAPY_DATA/cta-1dc/caldb/data/cta/1dc/bcf/South_z20_50h/irf_file.fits"
    )
    livetime = 0.1 * u.hr
    skydir = SkyCoord(0, 0, unit="deg", frame="galactic")
    pointing = FixedPointingInfo(fixed_icrs=skydir.icrs)
    obs = Observation.create(
        obs_id=1001,
        pointing=pointing,
        livetime=livetime,
        irfs=irfs,
        location=LOCATION,
    )

    energy_axis = MapAxis.from_energy_bounds(
        "1.0 TeV", "10 TeV", nbin=10, per_decade=True
    )
    energy_axis_true = MapAxis.from_energy_bounds(
        "0.5 TeV", "20 TeV", nbin=20, per_decade=True, name="energy_true"
    )
    migra_axis = MapAxis.from_bounds(0.5, 2, nbin=150, node_type="edges", name="migra")

    geom = WcsGeom.create(
        skydir=skydir,
        width=(2, 2),
        binsz=0.06,
        frame="icrs",
        axes=[energy_axis],
    )

    empty = MapDataset.create(
        geom,
        energy_axis_true=energy_axis_true,
        migra_axis=migra_axis,
        name="test",
    )
    maker = MapDatasetMaker(selection=["exposure", "background", "psf", "edisp"])
    dataset = maker.run(empty, obs)

    dataset.models = model_alternative
    sampler = MapDatasetEventSampler(random_state=0)
    events = sampler.run(dataset=dataset, observation=obs)

    assert len(events.table) == 215
    assert len(np.where(events.table["MC_ID"] == 0)[0]) == 40

    meta = events.table.meta
    assert meta["MID00000"] == 0
    assert meta["MMN00000"] == "test-bkg"
    assert meta["MID00001"] == 1
    assert meta["MID00002"] == 2
    assert meta["MID00003"] == 3
    assert meta["NMCIDS"] == 4


@requires_data()
def test_MC_ID_NMCID(model_alternative):
    irfs = load_irf_dict_from_file(
        "$GAMMAPY_DATA/cta-1dc/caldb/data/cta/1dc/bcf/South_z20_50h/irf_file.fits"
    )
    livetime = 0.1 * u.hr
    skydir = SkyCoord(0, 0, unit="deg", frame="galactic")
    pointing = FixedPointingInfo(fixed_icrs=skydir.icrs)
    obs = Observation.create(
        obs_id=1001,
        pointing=pointing,
        livetime=livetime,
        irfs=irfs,
        location=LOCATION,
    )

    energy_axis = MapAxis.from_energy_bounds(
        "1.0 TeV", "10 TeV", nbin=10, per_decade=True
    )
    energy_axis_true = MapAxis.from_energy_bounds(
        "0.5 TeV", "20 TeV", nbin=20, per_decade=True, name="energy_true"
    )
    migra_axis = MapAxis.from_bounds(0.5, 2, nbin=150, node_type="edges", name="migra")

    geom = WcsGeom.create(
        skydir=skydir,
        width=(2, 2),
        binsz=0.06,
        frame="icrs",
        axes=[energy_axis],
    )

    empty = MapDataset.create(
        geom,
        energy_axis_true=energy_axis_true,
        migra_axis=migra_axis,
        name="test",
    )
    maker = MapDatasetMaker(selection=["exposure", "background", "psf", "edisp"])
    dataset = maker.run(empty, obs)

    model_alternative[0].spectral_model.parameters["amplitude"].value = 1e-16
    dataset.models = model_alternative
    sampler = MapDatasetEventSampler(random_state=0)
    events = sampler.run(dataset=dataset, observation=obs)

    assert len(events.table) == 47
    assert len(np.where(events.table["MC_ID"] == 0)[0]) == 47

    meta = events.table.meta
    assert meta["MID00000"] == 0
    assert meta["MMN00000"] == "test-bkg"
    assert meta["MID00001"] == 1
    assert meta["MID00002"] == 2
    assert meta["MID00003"] == 3
    assert meta["NMCIDS"] == 4


@requires_data()
def test_MC_ID_flag(model_alternative):
    irfs = load_irf_dict_from_file(
        "$GAMMAPY_DATA/cta-1dc/caldb/data/cta/1dc/bcf/South_z20_50h/irf_file.fits"
    )
    livetime = 0.1 * u.hr
    skydir = SkyCoord(0, 0, unit="deg", frame="galactic")
    pointing = FixedPointingInfo(fixed_icrs=skydir.icrs)
    obs = Observation.create(
        obs_id=1001,
        pointing=pointing,
        livetime=livetime,
        irfs=irfs,
        location=LOCATION,
    )

    energy_axis = MapAxis.from_energy_bounds(
        "1.0 TeV", "10 TeV", nbin=10, per_decade=True
    )
    energy_axis_true = MapAxis.from_energy_bounds(
        "0.5 TeV", "20 TeV", nbin=20, per_decade=True, name="energy_true"
    )
    migra_axis = MapAxis.from_bounds(0.5, 2, nbin=150, node_type="edges", name="migra")

    geom = WcsGeom.create(
        skydir=skydir,
        width=(2, 2),
        binsz=0.06,
        frame="icrs",
        axes=[energy_axis],
    )

    empty = MapDataset.create(
        geom,
        energy_axis_true=energy_axis_true,
        migra_axis=migra_axis,
        name="test",
    )
    maker = MapDatasetMaker(selection=["exposure", "background", "psf", "edisp"])
    dataset = maker.run(empty, obs)

    model_alternative[0].spectral_model.parameters["amplitude"].value = 1e-16
    dataset.models = model_alternative
    sampler = MapDatasetEventSampler(random_state=0, keep_mc_id=False)
    events = sampler.run(dataset=dataset, observation=obs)

    meta = events.table.meta
    assert len(events.table) == 47
    assert "MC_ID" not in events.table.colnames
    assert "MID00000" not in meta.keys()
    assert "MMN00000" not in meta.keys()
    assert "MID00001" not in meta.keys()
    assert "MID00002" not in meta.keys()
    assert "MID00003" not in meta.keys()
    assert "NMCIDS" not in meta.keys()


@requires_data()
def test_bunch_event_number_sample_sources(dataset):
    spatial_model = GaussianSpatialModel(
        lon_0="0 deg", lat_0="0 deg", sigma="0.2 deg", frame="galactic"
    )
    spectral_model = PowerLawSpectralModel(amplitude="4e-10 cm-2 s-1 TeV-1")

    dataset.models = [
        SkyModel(spectral_model=spectral_model, spatial_model=spatial_model),
        FoVBackgroundModel(dataset_name=dataset.name),
    ]

    sampler = MapDatasetEventSampler(random_state=0, n_event_bunch=1000)
    events = sampler.run(dataset=dataset)

    assert len(events.table) == 24128


@requires_data()
def test_sort_evt_by_time(dataset):
    spatial_model = GaussianSpatialModel(
        lon_0="0 deg", lat_0="0 deg", sigma="0.2 deg", frame="galactic"
    )
    spectral_model = PowerLawSpectralModel(amplitude="4e-10 cm-2 s-1 TeV-1")

    dataset.models = [
        SkyModel(spectral_model=spectral_model, spatial_model=spatial_model),
        FoVBackgroundModel(dataset_name=dataset.name),
    ]

    sampler = MapDatasetEventSampler(random_state=0, n_event_bunch=1000)
    events = sampler.run(dataset=dataset)

    dt = events.table["TIME"][1:] - events.table["TIME"][:-1]
    assert np.all(dt.to("s") >= 0.0)


@requires_data()
def test_observation_event_sampler(signal_model):
    from gammapy.datasets.simulate import ObservationEventSampler

    datastore = DataStore.from_dir("$GAMMAPY_DATA/hess-dl3-dr1/")
    obs = datastore.get_observations()[0]

    # Use H.E.S.S. for testing otherwise the EdispMap computation takes too much time and memory with CTAO
    maker = ObservationEventSampler()

    sim_obs = maker.run(obs, None)
    assert sim_obs.events is not None
    assert len(sim_obs.events.table) > 0

    irfs = load_irf_dict_from_file(
        "$GAMMAPY_DATA/cta-caldb/Prod5-South-20deg-AverageAz-14MSTs37SSTs.180000s-v0.1.fits.gz"
    )
    pointing = FixedPointingInfo(
        fixed_icrs=SkyCoord(83.63311446, 22.01448714, unit="deg", frame="icrs"),
    )
    time_start = Time("2021-11-20T03:00:00")
    time_stop = Time("2021-11-20T03:30:00")

    obs = Observation.create(
        pointing=pointing,
        location=LOCATION,
        obs_id=1,
        tstart=time_start,
        tstop=time_stop,
        irfs=irfs,
        deadtime_fraction=0.01,
    )

    dataset_kwargs = dict(
        spatial_width=5 * u.deg,
        spatial_bin_size=0.01 * u.deg,
        energy_axis=MapAxis.from_energy_bounds(
            10 * u.GeV, 100 * u.TeV, nbin=5, per_decade=True
        ),
        energy_axis_true=MapAxis.from_energy_bounds(
            10 * u.GeV, 100 * u.TeV, nbin=5, per_decade=True, name="energy_true"
        ),
    )
    maker = ObservationEventSampler(dataset_kwargs=dataset_kwargs)

    sim_obs = maker.run(obs, [signal_model])
    assert sim_obs.events is not None
    assert len(sim_obs.events.table) > 0



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Simulate observations."""

import html
import logging
from copy import deepcopy
import numpy as np
import astropy.units as u
from astropy.coordinates import SkyCoord, SkyOffsetFrame
from astropy.table import Table, vstack
from astropy.time import Time
from gammapy import __version__
from gammapy.data import EventList, observatory_locations
from gammapy.maps import MapAxis, MapCoord, RegionNDMap, TimeMapAxis
from gammapy.modeling.models import (
    ConstantSpectralModel,
    ConstantTemporalModel,
    PointSpatialModel,
)
from gammapy.utils.fits import earth_location_to_dict
from gammapy.utils.random import get_random_state
from .map import create_map_dataset_from_observation

__all__ = ["MapDatasetEventSampler", "ObservationEventSampler"]

log = logging.getLogger(__name__)


class MapDatasetEventSampler:
    """Sample events from a map dataset.

    Parameters
    ----------
    random_state : {int, 'random-seed', 'global-rng', `~numpy.random.RandomState`}, optional
        Defines random number generator initialisation via the `~gammapy.utils.random.get_random_state` function.
    oversample_energy_factor : int, optional
        Defines an oversampling factor for the energies; it is used only when sampling
        an energy-dependent time-varying source.
    t_delta : `~astropy.units.Quantity`, optional
        Time interval used to sample the time-dependent source.
    keep_mc_id : bool, optional
        Flag to tag sampled events from a given model with a Montecarlo identifier.
        Default is True. If set to False, no identifier will be assigned.
    n_event_bunch : int
        Size of events bunches to sample. If None, sample all events in memory.
        Default is 10000.
    """

    def __init__(
        self,
        random_state="random-seed",
        oversample_energy_factor=10,
        t_delta=0.5 * u.s,
        keep_mc_id=True,
        n_event_bunch=10000,
    ):
        self.random_state = get_random_state(random_state)
        self.oversample_energy_factor = oversample_energy_factor
        self.t_delta = t_delta
        self.keep_mc_id = keep_mc_id
        self.n_event_bunch = n_event_bunch

    def _repr_html_(self):
        try:
            return self.to_html()
        except AttributeError:
            return f"<pre>{html.escape(str(self))}</pre>"

    def _make_table(self, coords):
        """Create a table for sampled events.

        Parameters
        ----------
        coords : `~gammapy.maps.MapCoord`
            Coordinates of the sampled events.

        Returns
        -------
        table : `~astropy.table.Table`
            Table of the sampled events.
        """
        table = Table()
        try:
            energy = coords["energy_true"]
        except KeyError:
            energy = coords["energy"]

        table["TIME"] = coords["time"]
        table["ENERGY_TRUE"] = energy

        table["RA_TRUE"] = coords.skycoord.icrs.ra.to("deg")
        table["DEC_TRUE"] = coords.skycoord.icrs.dec.to("deg")

        return table

    def _evaluate_timevar_source(
        self,
        dataset,
        model,
    ):
        """Calculate Npred for a given `dataset.model` by evaluating it on a region geometry.

        Parameters
        ----------
        dataset : `~gammapy.datasets.MapDataset`
            Map dataset.
        model : `~gammapy.modeling.models.SkyModel`
            Sky model instance.

        Returns
        -------
        npred : `~gammapy.maps.RegionNDMap`
            Npred map.
        """
        energy_true = dataset.edisp.edisp_map.geom.axes["energy_true"]
        energy_new = energy_true.upsample(self.oversample_energy_factor)

        position = model.spatial_model.position
        region_exposure = dataset.exposure.to_region_nd_map(position)

        time_axis_eval = TimeMapAxis.from_gti_bounds(
            gti=dataset.gti, t_delta=self.t_delta
        )

        time_min, time_max = time_axis_eval.time_bounds
        time_axis = MapAxis.from_bounds(
            time_min.mjd * u.d,
            time_max.mjd * u.d,
            nbin=time_axis_eval.nbin,
            name="time",
        )

        temp_eval = model.temporal_model.evaluate(
            time_axis_eval.time_mid, energy=energy_new.center
        )

        norm = model.spectral_model(energy=energy_new.center)

        if temp_eval.unit.is_equivalent(norm.unit):
            flux_diff = temp_eval.to(norm.unit) * norm.value[:, None]
        else:
            flux_diff = temp_eval * norm[:, None]

        flux_inte = flux_diff * energy_new.bin_width[:, None]

        flux_pred = RegionNDMap.create(
            region=position,
            axes=[time_axis, energy_new],
            data=np.array(flux_inte),
            unit=flux_inte.unit,
        )

        mapcoord = flux_pred.geom.get_coord(sparse=True)
        mapcoord["energy_true"] = energy_true.center[:, None, None, None]

        flux_values = flux_pred.interp_by_coord(mapcoord) * flux_pred.unit
        data = flux_values * region_exposure.quantity[:, None, :, :]
        data /= time_axis.nbin / self.oversample_energy_factor

        npred = RegionNDMap.create(
            region=position,
            axes=[time_axis, energy_true],
            data=data.to_value(""),
        )

        return npred

    def _sample_coord_time_energy(self, dataset, model):
        """Sample model components of a source with time-dependent spectrum.

        Parameters
        ----------
        dataset : `~gammapy.datasets.MapDataset`
            Map dataset.
        model : `~gammapy.modeling.models.SkyModel`
            Sky model instance.

        Returns
        -------
        table : `~astropy.table.Table`
            Table of sampled events.
        """
        if not isinstance(model.spatial_model, PointSpatialModel):
            raise TypeError(
                f"Event sampler expects PointSpatialModel for a time varying source. Got {model.spatial_model} instead."
            )

        if not isinstance(model.spectral_model, ConstantSpectralModel):
            raise TypeError(
                f"Event sampler expects ConstantSpectralModel for a time varying source. Got {model.spectral_model} instead."
            )

        npred = self._evaluate_timevar_source(dataset, model=model)
        data = npred.data[np.isfinite(npred.data)]
        data = np.clip(data, 0, None)

        try:
            n_events = self.random_state.poisson(np.sum(data))
        except ValueError:
            raise ValueError(
                f"The number of predicted events for the model {model.name} is too large. No event sampling will be performed for this model!"
            )

        coords = npred.sample_coord(n_events=n_events, random_state=self.random_state)

        coords["time"] = Time(
            coords["time"], format="mjd", scale=dataset.gti.time_ref.scale
        )

        table = self._make_table(coords)

        return table

    def _sample_coord_time(self, npred, temporal_model, gti):
        """Sample model components of a time-varying source.

        Parameters
        ----------
        npred : `~gammapy.dataset.MapDataset`
            Npred map.
        temporal_model : `~gammapy.modeling.models\
            Temporal model of the source.
        gti : `~gammapy.data.GTI`
             GTI of the dataset.

        Returns
        -------
        table : `~astropy.table.Table`
            Table of sampled events.
        """
        data = npred.data[np.isfinite(npred.data)]
        data = np.clip(data, 0, None)
        n_events = self.random_state.poisson(np.sum(data))

        coords = npred.sample_coord(n_events=n_events, random_state=self.random_state)

        time_start, time_stop = (gti.time_start, gti.time_stop)
        coords["time"] = temporal_model.sample_time(
            n_events=n_events,
            t_min=time_start,
            t_max=time_stop,
            random_state=self.random_state,
            t_delta=self.t_delta,
        )

        table = self._make_table(coords)

        return table

    def sample_sources(self, dataset, psf_update=False):
        """Sample source model components.

        Parameters
        ----------
        dataset : `~gammapy.datasets.MapDataset`
            Map dataset.
        psf_update : bool
            Parameter to switch-off (on) the update of the PSF
            in the dataset; default is False.

        Returns
        -------
        events : `~gammapy.data.EventList`
            Event list.
        """
        if psf_update is True:
            psf_update = dataset.psf
        else:
            psf_update = None

        events_all = Table()
        for idx, evaluator in enumerate(dataset.evaluators.values()):
            log.info(f"Evaluating model: {evaluator.model.name}")
            if evaluator.needs_update:
                evaluator.update(
                    dataset.exposure,
                    psf_update,
                    dataset.edisp,
                    dataset._geom,
                    dataset.mask,
                )
            if not evaluator.contributes:
                continue

            if evaluator.model.temporal_model is None:
                temporal_model = ConstantTemporalModel()
            else:
                temporal_model = evaluator.model.temporal_model

            if temporal_model.is_energy_dependent:
                table = self._sample_coord_time_energy(dataset, evaluator.model)
            else:
                flux = evaluator.compute_flux()
                npred = evaluator.apply_exposure(flux)
                table = self._sample_coord_time(npred, temporal_model, dataset.gti)

            if self.keep_mc_id:
                if len(table) == 0:
                    mcid = table.Column(name="MC_ID", length=0, dtype=int)
                    table.add_column(mcid)

                table["MC_ID"] = idx + 1
                table.meta["MID{:05d}".format(idx + 1)] = idx + 1
                table.meta["MMN{:05d}".format(idx + 1)] = evaluator.model.name

            events_all = vstack([events_all, table])

        return events_all

    def sample_background(self, dataset):
        """Sample background.

        Parameters
        ----------
        dataset : `~gammapy.datasets.MapDataset`
            Map dataset.

        Returns
        -------
        events : `gammapy.data.EventList`
            Background events.
        """
        table = Table()
        if dataset.background:
            log.info("Evaluating background...")
            background = dataset.npred_background()

            temporal_model = ConstantTemporalModel()

            table = self._sample_coord_time(background, temporal_model, dataset.gti)

            table["ENERGY"] = table["ENERGY_TRUE"]
            table["RA"] = table["RA_TRUE"]
            table["DEC"] = table["DEC_TRUE"]

            if self.keep_mc_id:
                table["MC_ID"] = 0
                table.meta["MID{:05d}".format(0)] = 0
                table.meta["MMN{:05d}".format(0)] = dataset.background_model.name

        return table

    def sample_edisp(self, edisp_map, events_table):
        """Sample energy dispersion map.

        Parameters
        ----------
        edisp_map : `~gammapy.irf.EDispMap`
            Energy dispersion map.
        events_table : `~astropy.table.Table`
            Event list with the true energies.

        Returns
        -------
        events_table : `~astropy.table.Table`
            Event list with reconstructed energy column.
        """
        coord = MapCoord(
            {
                "lon": events_table["RA_TRUE"].quantity,
                "lat": events_table["DEC_TRUE"].quantity,
                "energy_true": events_table["ENERGY_TRUE"].quantity,
            },
            frame="icrs",
        )

        coords_reco = edisp_map.sample_coord(
            coord, self.random_state, self.n_event_bunch
        )
        events_table["ENERGY"] = coords_reco["energy"]
        return events_table

    def sample_psf(self, psf_map, events_table):
        """Sample PSF map.

        Parameters
        ----------
        psf_map : `~gammapy.irf.PSFMap`
            PSF map.
        events : `~gammapy.data.EventList`
            Event list.

        Returns
        -------
        events_table : `~astropy.table.Table`
            Event list with reconstructed position columns.
        """
        coord = MapCoord(
            {
                "lon": events_table["RA_TRUE"].quantity,
                "lat": events_table["DEC_TRUE"].quantity,
                "energy_true": events_table["ENERGY_TRUE"].quantity,
            },
            frame="icrs",
        )

        coords_reco = psf_map.sample_coord(coord, self.random_state, self.n_event_bunch)

        events_table["RA"] = coords_reco["lon"] * u.deg
        events_table["DEC"] = coords_reco["lat"] * u.deg
        return events_table

    @staticmethod
    def event_det_coords(observation, events):
        """Add columns of detector coordinates (DETX-DETY) to the event list.

        Parameters
        ----------
        observation : `~gammapy.data.Observation`
            In memory observation.
        events : `~gammapy.data.EventList`
            Event list.

        Returns
        -------
        events : `~gammapy.data.EventList`
            Event list with columns of event detector coordinates.
        """
        sky_coord = SkyCoord(events.table["RA"], events.table["DEC"], frame="icrs")
        frame = SkyOffsetFrame(origin=observation.get_pointing_icrs(observation.tmid))
        pseudo_fov_coord = sky_coord.transform_to(frame)

        events.table["DETX"] = pseudo_fov_coord.lon
        events.table["DETY"] = pseudo_fov_coord.lat
        return events

    @staticmethod
    def event_list_meta(dataset, observation, keep_mc_id=True):
        """Event list meta info.

        Please, note that this function will be updated in the future.

        Parameters
        ----------
        dataset : `~gammapy.datasets.MapDataset`
            Map dataset.
        observation : `~gammapy.data.Observation`
            In memory observation.
        keep_mc_id : bool
            Flag to tag sampled events from a given model with a Montecarlo identifier.
            Default is True. If set to False, no identifier will be assigned.

        Returns
        -------
        meta : dict
            Meta dictionary.
        """
        # See: https://gamma-astro-data-formats.readthedocs.io/en/latest/events/events.html#mandatory-header-keywords  # noqa: E501
        meta = {}

        meta["HDUCLAS1"] = "EVENTS"
        meta["EXTNAME"] = "EVENTS"
        meta["HDUDOC"] = (
            "https://github.com/open-gamma-ray-astro/gamma-astro-data-formats"
        )
        meta["HDUVERS"] = "0.2"
        meta["HDUCLASS"] = "GADF"

        meta["OBS_ID"] = observation.obs_id

        meta["TSTART"] = (observation.tstart - dataset.gti.time_ref).to_value("s")
        meta["TSTOP"] = (observation.tstop - dataset.gti.time_ref).to_value("s")

        meta["ONTIME"] = observation.observation_time_duration.to("s").value
        meta["LIVETIME"] = observation.observation_live_time_duration.to("s").value
        meta["DEADC"] = 1 - observation.observation_dead_time_fraction

        fixed_icrs = observation.pointing.fixed_icrs
        meta["RA_PNT"] = fixed_icrs.ra.deg
        meta["DEC_PNT"] = fixed_icrs.dec.deg

        meta["EQUINOX"] = "J2000"
        meta["RADECSYS"] = "icrs"

        meta["CREATOR"] = "Gammapy {}".format(__version__)
        meta["EUNIT"] = "TeV"
        meta["EVTVER"] = ""

        meta["OBSERVER"] = "Gammapy user"
        meta["DSTYP1"] = "TIME"
        meta["DSUNI1"] = "s"
        meta["DSVAL1"] = "TABLE"
        meta["DSREF1"] = ":GTI"
        meta["DSTYP2"] = "ENERGY"
        meta["DSUNI2"] = "TeV"
        meta["DSVAL2"] = (
            f'{dataset._geom.axes["energy"].edges.min().value}:{dataset._geom.axes["energy"].edges.max().value}'  # noqa: E501
        )
        meta["DSTYP3"] = "POS(RA,DEC)     "

        offset_max = np.max(dataset._geom.width).to_value("deg")
        meta["DSVAL3"] = (
            f"CIRCLE({fixed_icrs.ra.deg},{fixed_icrs.dec.deg},{offset_max})"  # noqa: E501
        )
        meta["DSUNI3"] = "deg             "
        meta["NDSKEYS"] = " 3 "

        # get first non background model component
        for model in dataset.models:
            if model is not dataset.background_model:
                break
        else:
            model = None

        if model:
            meta["OBJECT"] = model.name
            meta["RA_OBJ"] = model.position.icrs.ra.deg
            meta["DEC_OBJ"] = model.position.icrs.dec.deg

        meta["TELAPSE"] = dataset.gti.time_sum.to("s").value
        meta["MJDREFI"] = int(dataset.gti.time_ref.mjd)
        meta["MJDREFF"] = dataset.gti.time_ref.mjd % 1
        meta["TIMEUNIT"] = "s"
        meta["TIMESYS"] = dataset.gti.time_ref.scale
        meta["TIMEREF"] = "LOCAL"
        meta["DATE-OBS"] = dataset.gti.time_start.isot[0][0:10]
        meta["DATE-END"] = dataset.gti.time_stop.isot[0][0:10]
        meta["CONV_DEP"] = 0
        meta["CONV_RA"] = 0
        meta["CONV_DEC"] = 0

        if keep_mc_id:
            meta["NMCIDS"] = len(dataset.models)

        # Necessary for DataStore, but they should be ALT and AZ instead!
        telescope = observation.aeff.meta["TELESCOP"]
        instrument = observation.aeff.meta["INSTRUME"]
        loc = observation.observatory_earth_location
        if loc is None:
            if telescope == "CTA":
                if instrument == "Southern Array":
                    loc = observatory_locations["ctao_south"]
                elif instrument == "Northern Array":
                    loc = observatory_locations["ctao_north"]
                else:
                    loc = observatory_locations["ctao_south"]

            else:
                loc = observatory_locations[telescope.lower()]

        meta.update(earth_location_to_dict(loc))

        # this is not really correct but maybe OK for now
        coord_altaz = observation.pointing.get_altaz(dataset.gti.time_start, loc)

        meta["ALT_PNT"] = coord_altaz.alt.deg[0]
        meta["AZ_PNT"] = coord_altaz.az.deg[0]

        # TO DO: these keywords should be taken from the IRF of the dataset
        meta["ORIGIN"] = "Gammapy"
        meta["TELESCOP"] = observation.aeff.meta["TELESCOP"]
        meta["INSTRUME"] = observation.aeff.meta["INSTRUME"]
        meta["N_TELS"] = ""
        meta["TELLIST"] = ""

        meta["CREATED"] = Time.now().iso
        meta["OBS_MODE"] = "POINTING"
        meta["EV_CLASS"] = ""

        return meta

    def run(self, dataset, observation=None):
        """Run the event sampler, applying IRF corrections.

        Parameters
        ----------
        dataset : `~gammapy.datasets.MapDataset`
            Map dataset.
        observation : `~gammapy.data.Observation`, optional
            In memory observation. Default is None.

        Returns
        -------
        events : `~gammapy.data.EventList`
            Event list.
        """
        events_src = self.sample_sources(dataset)

        if len(events_src) > 0:
            if dataset.psf:
                events_src = self.sample_psf(dataset.psf, events_src)
            else:
                events_src["RA"] = events_src["RA_TRUE"]
                events_src["DEC"] = events_src["DEC_TRUE"]

            if dataset.edisp:
                events_src = self.sample_edisp(dataset.edisp, events_src)
            else:
                events_src["ENERGY"] = events_src["ENERGY_TRUE"]

        events_bkg = self.sample_background(dataset)
        events = vstack([events_src, events_bkg])
        events = EventList(events)

        events.table["EVENT_ID"] = np.arange(len(events.table))
        if observation is not None:
            events = self.event_det_coords(observation, events)
            events.table.meta.update(
                self.event_list_meta(dataset, observation, self.keep_mc_id)
            )

        sort_by_time = np.argsort(events.table["TIME"])
        events.table = events.table[sort_by_time]

        geom = dataset._geom
        selection = geom.contains(events.map_coord(geom))
        log.info("Event sampling completed.")
        return events.select_row_subset(selection)


class ObservationEventSampler(MapDatasetEventSampler):
    """
    Sample event lists for a given observation and signal models.

    Signal events are sampled from the predicted counts distribution given by the product of the sky models and the
    expected exposure. They are then folded with the instrument response functions.
    To improve performance, IRFs are evaluated on a pre-defined binning,
    not at each individual event energy / coordinate.

    Parameters
    ----------
    random_state : {int, 'random-seed', 'global-rng', `~numpy.random.RandomState`}, optional
        Defines random number generator initialisation via the `~gammapy.utils.random.get_random_state` function.
    oversample_energy_factor : int, optional
        Defines an oversampling factor for the energies; it is used only when sampling
        an energy-dependent time-varying source. Default is 10.
    t_delta : `~astropy.units.Quantity`, optional
        Time interval used to sample the time-dependent source. Default is 0.5 s.
    keep_mc_id : bool, optional
        Flag to tag sampled events from a given model with a Montecarlo identifier.
        Default is True. If set to False, no identifier will be assigned.
    n_event_bunch : int
        Size of events bunches to sample. If None, sample all events in memory.
        Default is 10000.
    dataset_kwargs : dict, optional
        Arguments passed to `~gammapy.datasets.create_map_dataset_from_observation()`
    """

    def __init__(
        self,
        random_state="random-seed",
        oversample_energy_factor=10,
        t_delta=0.5 * u.s,
        keep_mc_id=True,
        n_event_bunch=10000,
        dataset_kwargs=None,
    ):
        self.dataset_kwargs = dataset_kwargs or {}
        self.random_state = get_random_state(random_state)
        self.oversample_energy_factor = oversample_energy_factor
        self.t_delta = t_delta
        self.n_event_bunch = n_event_bunch
        self.keep_mc_id = keep_mc_id

    def run(self, observation, models=None, dataset_name=None):
        """Sample events for given observation and signal models.

        The signal distribution is sampled from the given models
        in true coordinates and energy. The true quantities are
        then folded with the IRFs to obtain the observable quantities.

        Parameters
        ----------
        observation : `~gammapy.data.Observation`
            Observation to be simulated.
        models : `~gammapy.modeling.Models`, optional
            Models to simulate.
            Can be None to only sample background events. Default is None.
        dataset_name : str, optional
            If `models` contains one or multiple `FoVBackgroundModel`
            it should match the `dataset_name` of the background model to use.
            Default is None.

        Returns
        -------
        observation : `~gammapy.data.Observation`
            A copy of the input observation with event list filled.
        """
        dataset = create_map_dataset_from_observation(
            observation, models, dataset_name, **self.dataset_kwargs
        )
        events = super().run(dataset, observation)

        observation = deepcopy(observation)
        observation._events = events
        return observation



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import sys
import pytest
import logging
import numpy as np
from numpy.testing import assert_allclose
import astropy.units as u
from astropy.coordinates import SkyCoord
from gammapy.catalog import SourceCatalog4FGL
from gammapy.maps import MapAxis, WcsGeom
from gammapy.modeling import Parameter, Parameters
from gammapy.modeling.models import (
    GaussianSpatialModel,
    GeneralizedGaussianSpatialModel,
    Shell2SpatialModel,
    ShellSpatialModel,
    Model,
    ModelBase,
    Models,
    PointSpatialModel,
    PowerLawSpectralModel,
    SkyModel,
    FoVBackgroundModel,
    DiskSpatialModel,
)
from gammapy.utils.testing import mpl_plot_check, requires_data


class MyModel(ModelBase):
    """Simple model example"""

    x = Parameter("x", 1, "cm")
    y = Parameter("y", 2)


class CoModel(ModelBase):
    """Compound model example"""

    norm = Parameter("norm", 42, "cm")

    def __init__(self, m1, m2, norm=norm.quantity):
        self.m1 = m1
        self.m2 = m2
        super().__init__(norm=norm)

    @property
    def parameters(self):
        return Parameters([self.norm]) + self.m1.parameters + self.m2.parameters


class WrapperModel(ModelBase):
    """Wrapper compound model.

    Dynamically generated parameters in `__init__`,
    and a parameter name conflict with the wrapped
    model, both have a parameter called "y".
    """

    def __init__(self, m1, a=1, y=99):
        self.m1 = m1
        a = Parameter("a", a)
        y = Parameter("y", y)
        self.default_parameters = Parameters([a, y])
        super().__init__(a=a, y=y)

    @property
    def parameters(self):
        return Parameters([self.a, self.y]) + self.m1.parameters


def test_model_class():
    assert isinstance(MyModel.parameters, property)
    assert MyModel.x.name == "x"
    assert MyModel.default_parameters["x"] is MyModel.x


def test_model_class_par_init():
    x = Parameter("x", 4, "cm")
    y = Parameter("y", 10)

    model = MyModel(x=x, y=y)

    assert x is model.x
    assert y is model.y


def test_model_init():
    m = MyModel()
    assert m.x.name == "x"
    assert m.x.value == 1
    assert m.x is m.parameters[0]
    assert m.y is m.parameters[1]
    assert m.parameters is not MyModel.default_parameters

    m = MyModel(x="99 cm")
    assert m.x.value == 99
    assert m.y.value == 2

    # Currently we always convert to the default unit of a parameter
    # TODO: discuss if this is the behaviour we want, or if we instead
    # should change to the user-set unit, as long as it's compatible
    m = MyModel(x=99 * u.m)
    assert_allclose(m.x.value, 99)
    assert m.x.unit == "m"

    with pytest.raises(u.UnitConversionError):
        MyModel(x=99)

    with pytest.raises(u.UnitConversionError):
        MyModel(x=99 * u.s)


def test_wrapper_model():
    outer = MyModel()
    m = WrapperModel(outer)

    assert isinstance(m.a, Parameter)
    assert m.y.value == 99

    assert m.parameters.names == ["a", "y", "x", "y"]


def test_model_parameter():
    m = MyModel(x="99 cm")
    assert isinstance(m.x, Parameter)
    assert m.x.value == 99
    assert m.x.unit == "cm"

    with pytest.raises(TypeError):
        m.x = 99

    with pytest.raises(TypeError):
        m.x = 99 * u.cm


# TODO: implement parameter linking. Not working ATM!
def test_model_parameter_link():
    # Assigning a parameter should create a link
    m = MyModel()
    par = MyModel.x.copy()
    m.x = par
    assert isinstance(m.x, Parameter)
    assert m.x is par
    # model.parameters should be in sync with attributes
    assert m.x is m.parameters["x"]


def test_model_copy():
    m = MyModel()

    m2 = m.copy()

    # Models should be independent
    assert m.parameters is not m2.parameters
    assert m.parameters[0] is not m2.parameters[0]


def test_model_create():
    spectral_model = Model.create(
        "pl-2", model_type="spectral", amplitude="1e-10 cm-2 s-1", index=3
    )
    assert "PowerLaw2SpectralModel" in spectral_model.tag
    assert_allclose(spectral_model.index.value, 3)


def test_compound_model():
    m1 = MyModel()
    m2 = MyModel(x=10 * u.cm, y=20)
    m = CoModel(m1, m2)
    assert len(m.parameters) == 5
    assert m.parameters.names == ["norm", "x", "y", "x", "y"]
    assert_allclose(m.parameters.value, [42, 1, 2, 10, 20])


def test_parameter_link_init():
    m1 = MyModel()
    m2 = MyModel(y=m1.y)

    assert m1.y is m2.y

    m1.y.value = 100
    assert_allclose(m2.y.value, 100)


def test_parameter_link():
    m1 = MyModel()
    m2 = MyModel()

    m2.y = m1.y

    m1.y.value = 100
    assert_allclose(m2.y.value, 100)


@requires_data()
def test_set_parameters_from_table():
    # read gammapy models
    models = Models.read("$GAMMAPY_DATA/tests/models/gc_example_models.yaml")

    tab = models.to_parameters_table()
    tab["value"][0] = 3.0
    tab["min"][0] = -10
    tab["max"][0] = 10
    tab["frozen"][0] = True
    tab["name"][0] = "index2"
    tab["frozen"][1] = True

    models.update_parameters_from_table(tab)

    d = models.parameters.to_dict()
    assert d[0]["value"] == 3.0
    assert d[0]["min"] == -10
    assert d[0]["max"] == 10
    assert d[0]["frozen"]
    assert d[0]["name"] == "index"

    assert d[1]["frozen"]


@requires_data()
def test_plot_models(caplog):
    models = Models.read("$GAMMAPY_DATA/tests/models/gc_example_models.yaml")

    with mpl_plot_check():
        models.plot_regions(linewidth=2)
        models.plot_regions()

    assert models.wcs_geom.data_shape == models.wcs_geom.wcs.array_shape

    regions = models.to_regions()
    assert len(regions) == 3

    p1 = Model.create(
        "pl-2",
        model_type="spectral",
    )
    g1 = Model.create("gauss", model_type="spatial")
    p2 = Model.create(
        "pl-2",
        model_type="spectral",
    )
    m1 = SkyModel(spectral_model=p1, spatial_model=g1, name="m1")
    m2 = SkyModel(spectral_model=p2, name="m2")
    models = Models([m1, m2])

    models.plot_regions()
    assert "WARNING" in [_.levelname for _ in caplog.records]
    assert "Skipping model m2 - no spatial component present" in [
        _.message for _ in caplog.records
    ]


def test_plot_models_empty(caplog):
    models = Models([])
    models.plot_regions()


def test_positions():
    p1 = Model.create(
        "pl",
        model_type="spectral",
    )
    g1 = Model.create("gauss", model_type="spatial")
    m1 = SkyModel(spectral_model=p1, spatial_model=g1, name="m1")
    g3 = Model.create("gauss", model_type="spatial", frame="galactic")
    m3 = SkyModel(spectral_model=p1, spatial_model=g3, name="m3")
    models = Models([m1, m3])
    pos = models.positions
    assert_allclose(pos.galactic[0].l.value, 96.337, rtol=1e-3)


def test_parameter_name():
    # From the 3.12 changelog:
    # Exceptions raised in a class or type’s __set_name__ method are no longer
    # wrapped by a RuntimeError.
    if sys.version_info < (3, 12):
        exc_class = RuntimeError
    else:
        exc_class = ValueError

    with pytest.raises(exc_class):

        class MyTestModel:
            par = Parameter("wrong-name", value=3)

        _ = MyTestModel()


@requires_data()
def test_select_models():
    cat = SourceCatalog4FGL()
    mask_models = cat.table["GLAT"].quantity > 80 * u.deg
    subcat = cat[mask_models]
    models = subcat.to_models()
    pos = SkyCoord(182, 25, unit="deg", frame="icrs")
    geom = WcsGeom.create(skydir=pos, width=2 * u.deg, binsz=0.02, frame="icrs")
    models_selected = models.select_from_geom(geom)
    assert len(models_selected) == 2


def test_to_template():
    energy_bounds = [1, 100] * u.TeV
    energy_axis = MapAxis.from_energy_bounds(
        energy_bounds[0], energy_bounds[1], nbin=2, per_decade=True, name="energy_true"
    )

    spatial_model = GaussianSpatialModel()
    spectral_model = PowerLawSpectralModel()
    geom = spatial_model._evaluation_geom.to_cube([energy_axis])

    model = SkyModel(
        spatial_model=spatial_model, spectral_model=PowerLawSpectralModel()
    )
    models = Models([model])

    template3d = models.to_template_sky_model(geom)

    template_1d_direct = models.to_template_spectral_model(geom)
    template_1d_from3d = Models([template3d]).to_template_spectral_model(geom)

    energy_axis_down = energy_axis.upsample(2)
    values_ref = spectral_model(energy_axis_down.edges)
    values_direct = template_1d_direct(energy_axis_down.edges)
    values_from3d = template_1d_from3d(energy_axis_down.edges)

    assert_allclose(values_ref, values_direct, rtol=1e-5)
    assert_allclose(values_ref, values_from3d, rtol=1e-5)


def test_add_not_unique_models():
    spec_model1 = PowerLawSpectralModel()
    spatial_model1 = PointSpatialModel()

    model1 = SkyModel(
        spectral_model=spec_model1, spatial_model=spatial_model1, name="source1"
    )

    model2 = SkyModel(
        spectral_model=spec_model1, spatial_model=spatial_model1, name="source2"
    )

    model3 = SkyModel(
        spectral_model=spec_model1, spatial_model=spatial_model1, name="source3"
    )

    model4 = SkyModel(
        spectral_model=spec_model1, spatial_model=spatial_model1, name="source1"
    )

    models1 = Models([model1, model2])
    models2 = Models([model3, model4])

    with pytest.raises(
        ValueError,
        match="Model names must be unique. Models named 'source1' are duplicated.",
    ):
        models1.extend(models2)


def test_two_fov_bkg_models_single_dataset():
    fov1 = FoVBackgroundModel(dataset_name="ds1", name="ds1-1")
    fov2 = FoVBackgroundModel(dataset_name="ds1", name="ds1-2")
    with pytest.raises(
        ValueError,
        match="Only one FoVBackgroundModel per Dataset is permitted - already got one for ds1",
    ):
        Models([fov1, fov2])


def test_to_regions():
    spectral = PowerLawSpectralModel()
    spatials = [
        PointSpatialModel(),
        GaussianSpatialModel(sigma="1deg"),
        DiskSpatialModel(r_0="1deg"),
        GeneralizedGaussianSpatialModel(r_0="1deg"),
        ShellSpatialModel(
            radius="0.5deg",
            width="0.5deg",
        ),
        Shell2SpatialModel(r_0="1deg", eta=0.5),
    ]
    models = Models([SkyModel(spectral, spatial) for spatial in spatials])

    regs1 = models.to_regions()
    regs2 = models.to_regions(size_factor=3)
    assert len(regs1) == 6
    for reg1, reg2 in zip(regs1[1:], regs2[1:]):
        try:
            assert_allclose(reg1.width, 2 * u.deg, rtol=1e-5)
            assert_allclose(reg2.width, 6 * u.deg, rtol=1e-5)
        except AttributeError:
            assert_allclose(reg1.inner_radius, 0.5 * u.deg, rtol=1e-5)
            assert_allclose(reg2.inner_radius, 1.5 * u.deg, rtol=1e-5)
            assert_allclose(reg1.outer_radius, 1 * u.deg, rtol=1e-5)
            assert_allclose(reg2.outer_radius, 3 * u.deg, rtol=1e-5)


def test_bad_covariance(caplog, tmp_path):
    # Step 1: Create a simple spectral model
    spectral_model = PowerLawSpectralModel()
    model = SkyModel(spectral_model=spectral_model, name="test-model")
    models = Models([model])

    # Step 2: Assign a valid 3x3 identity covariance matrix
    models.covariance = np.identity(3)

    # Step 3: Write models.yaml and covariance.dat
    models_file = tmp_path / "models.yaml"
    models.write(models_file, write_covariance=True, overwrite=True)

    # Step 4: Overwrite covariance.dat with an invalid 1x1 matrix
    covariance_file = tmp_path / "models_covariance.dat"
    with open(covariance_file, "w") as f:
        f.write("|                    Parameters |   0 |\n")
        f.write("|     test-model.spectral.index | 1.0 |\n")

    # Step 5: Attempt to read the model and catch the expected error
    with caplog.at_level(logging.WARNING):
        Models.read(models_file)
        assert (
            "Impossible to read the covariance correctly" in caplog.records[0].message
        )



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Common fit statistics used in gamma-ray astronomy.

see :ref:`fit-statistics`
"""

from abc import ABC
import numpy as np
from scipy.special import erfc
from gammapy.maps import Map

from gammapy.utils.compilation import get_fit_statistics_compiled


__all__ = [
    "cash",
    "cstat",
    "wstat",
    "get_wstat_mu_bkg",
    "get_wstat_gof_terms",
    "CashFitStatistic",
    "WStatFitStatistic",
    "Chi2FitStatistic",
    "Chi2AsymmetricErrorFitStatistic",
]


def cash(n_on, mu_on, truncation_value=None):
    r"""Cash statistic, for Poisson data.

    The Cash statistic is defined as:

    .. math::
        C = 2 \left( \mu_{on} - n_{on} \log \mu_{on} \right)

    and :math:`C = 0` where :math:`\mu <= 0`.
    For more information see :ref:`fit-statistics`.

    Parameters
    ----------
    n_on : `~numpy.ndarray` or array_like
        Observed counts.
    mu_on : `~numpy.ndarray` or array_like
        Expected counts.
    truncation_value : `~numpy.ndarray` or array_like
        Minimum value use for ``mu_on``
        ``mu_on`` = ``truncation_value`` where ``mu_on`` <= ``truncation_value``.
        Default is 1e-25.

    Returns
    -------
    stat : ndarray
        Statistic per bin.

    References
    ----------
    * `Sherpa statistics page section on the Cash statistic
      <http://cxc.cfa.harvard.edu/sherpa/statistics/#cash>`_
    * `Sherpa help page on the Cash statistic
      <http://cxc.harvard.edu/sherpa/ahelp/cash.html>`_
    * `Cash (1979), ApJ 228, 939,
      <https://ui.adsabs.harvard.edu/abs/1979ApJ...228..939C>`_
    """
    if truncation_value is None:
        truncation_value = get_fit_statistics_compiled()["TRUNCATION_VALUE"]

    n_on = np.asanyarray(n_on)
    mu_on = np.asanyarray(mu_on)
    truncation_value = np.asanyarray(truncation_value)
    if np.any(truncation_value) <= 0:
        raise ValueError("Cash statistic truncation value must be positive.")

    mu_on = np.where(mu_on <= truncation_value, truncation_value, mu_on)

    # suppress zero division warnings, they are corrected below
    with np.errstate(divide="ignore", invalid="ignore"):
        stat = 2 * (mu_on - n_on * np.log(mu_on))
    return stat


def cstat(n_on, mu_on, truncation_value=None):
    r"""C statistic, for Poisson data.

    The C statistic is defined as:

    .. math::
        C = 2 \left[ \mu_{on} - n_{on} + n_{on}
            (\log(n_{on}) - log(\mu_{on}) \right]

    and :math:`C = 0` where :math:`\mu_{on} <= 0`.

    ``truncation_value`` handles the case where ``n_on`` or ``mu_on`` is 0 or less and
    the log cannot be taken.
    For more information see :ref:`fit-statistics`.

    Parameters
    ----------
    n_on : `~numpy.ndarray` or array_like
        Observed counts.
    mu_on : `~numpy.ndarray` or array_like
        Expected counts.
    truncation_value : float
        ``n_on`` = ``truncation_value`` where ``n_on`` <= ``truncation_value.``
        ``mu_on`` = ``truncation_value`` where ``n_on`` <= ``truncation_value``
        Default is 1e-25.

    Returns
    -------
    stat : ndarray
        Statistic per bin.

    References
    ----------
    * `Sherpa stats page section on the C statistic
      <http://cxc.cfa.harvard.edu/sherpa/statistics/#cstat>`_
    * `Sherpa help page on the C statistic
      <http://cxc.harvard.edu/sherpa/ahelp/cash.html>`_
    * `Cash (1979), ApJ 228, 939
      <https://ui.adsabs.harvard.edu/abs/1979ApJ...228..939C>`_
    """
    if truncation_value is None:
        truncation_value = get_fit_statistics_compiled()["TRUNCATION_VALUE"]

    n_on = np.asanyarray(n_on, dtype=np.float64)
    mu_on = np.asanyarray(mu_on, dtype=np.float64)
    truncation_value = np.asanyarray(truncation_value, dtype=np.float64)

    if np.any(truncation_value) <= 0:
        raise ValueError("Cstat statistic truncation value must be positive.")

    n_on = np.where(n_on <= truncation_value, truncation_value, n_on)
    mu_on = np.where(mu_on <= truncation_value, truncation_value, mu_on)

    term1 = np.log(n_on) - np.log(mu_on)
    stat = 2 * (mu_on - n_on + n_on * term1)
    stat = np.where(mu_on > 0, stat, 0)

    return stat


def wstat(n_on, n_off, alpha, mu_sig, mu_bkg=None, extra_terms=True):
    r"""W statistic, for Poisson data with Poisson background.

    For a definition of WStat see :ref:`wstat`. If ``mu_bkg`` is not provided
    it will be calculated according to the profile likelihood formula.

    Parameters
    ----------
    n_on : `~numpy.ndarray` or array_like
        Total observed counts.
    n_off : `~numpy.ndarray` or array_like
        Total observed background counts.
    alpha : `~numpy.ndarray` or array_like
        Exposure ratio between on and off region.
    mu_sig : `~numpy.ndarray` or array_like
        Signal expected counts.
    mu_bkg : `~numpy.ndarray` or array_like, optional
        Background expected counts.
    extra_terms : bool, optional
        Add model independent terms to convert stat into goodness-of-fit
        parameter. Default is True.

    Returns
    -------
    stat : ndarray
        Statistic per bin.

    References
    ----------
    * `Habilitation M. de Naurois, p. 141
      <http://inspirehep.net/record/1122589/files/these_short.pdf>`_
    * `XSPEC page on Poisson data with Poisson background
      <https://heasarc.gsfc.nasa.gov/xanadu/xspec/manual/XSappendixStatistics.html>`_
    """
    # Note: This is equivalent to what's defined on the XSPEC page under the
    # following assumptions
    # t_s * m_i = mu_sig
    # t_b * m_b = mu_bkg
    # t_s / t_b = alpha

    n_on = np.asanyarray(n_on, dtype=np.float64)
    n_off = np.asanyarray(n_off, dtype=np.float64)
    alpha = np.asanyarray(alpha, dtype=np.float64)
    mu_sig = np.asanyarray(mu_sig, dtype=np.float64)

    if mu_bkg is None:
        mu_bkg = get_wstat_mu_bkg(n_on, n_off, alpha, mu_sig)

    term1 = mu_sig + (1 + alpha) * mu_bkg

    # suppress zero division warnings, they are corrected below
    with np.errstate(divide="ignore", invalid="ignore"):
        # This is a false positive error from pylint
        # See https://github.com/PyCQA/pylint/issues/2436
        term2_ = -n_on * np.log(mu_sig + alpha * mu_bkg)  # pylint:disable=invalid-unary-operand-type
    # Handle n_on == 0
    condition = n_on == 0
    term2 = np.where(condition, 0, term2_)

    # suppress zero division warnings, they are corrected below
    with np.errstate(divide="ignore", invalid="ignore"):
        # This is a false positive error from pylint
        # See https://github.com/PyCQA/pylint/issues/2436
        term3_ = -n_off * np.log(mu_bkg)  # pylint:disable=invalid-unary-operand-type
    # Handle n_off == 0
    condition = n_off == 0
    term3 = np.where(condition, 0, term3_)

    stat = 2 * (term1 + term2 + term3)

    if extra_terms:
        stat += get_wstat_gof_terms(n_on, n_off)

    return stat


def get_wstat_mu_bkg(n_on, n_off, alpha, mu_sig):
    """Background estimate ``mu_bkg`` for WSTAT.

    See :ref:`wstat`.
    """
    n_on = np.asanyarray(n_on, dtype=np.float64)
    n_off = np.asanyarray(n_off, dtype=np.float64)
    alpha = np.asanyarray(alpha, dtype=np.float64)
    mu_sig = np.asanyarray(mu_sig, dtype=np.float64)

    # NOTE: Corner cases in the docs are all handled correctly by this formula
    C = alpha * (n_on + n_off) - (1 + alpha) * mu_sig
    D = np.sqrt(C**2 + 4 * alpha * (alpha + 1) * n_off * mu_sig)
    with np.errstate(invalid="ignore", divide="ignore"):
        mu_bkg = (C + D) / (2 * alpha * (alpha + 1))

    return mu_bkg


def get_wstat_gof_terms(n_on, n_off):
    """Goodness of fit terms for WSTAT.

    See :ref:`wstat`.
    """
    term = np.zeros(n_on.shape)

    # suppress zero division warnings, they are corrected below
    with np.errstate(divide="ignore", invalid="ignore"):
        term1 = -n_on * (1 - np.log(n_on))
        term2 = -n_off * (1 - np.log(n_off))

    term += np.where(n_on == 0, 0, term1)
    term += np.where(n_off == 0, 0, term2)

    return 2 * term


class FitStatistic(ABC):
    """Abstract base class for FitStatistic objects."""

    @classmethod
    def stat_sum_dataset(cls, dataset):
        """Calculate -2 * sum log(L)."""
        stat_array = cls.stat_array_dataset(dataset)
        if dataset.mask is not None:
            mask = dataset.mask.data if isinstance(dataset.mask, Map) else dataset.mask
            stat_array = stat_array[mask]
        return np.sum(stat_array)

    @classmethod
    def stat_array_dataset(cls, dataset):
        """Calculate -2 * log(L)."""
        raise NotImplementedError

    @classmethod
    def loglikelihood_dataset(cls, dataset):
        """Calculate sum log(L)."""
        return -0.5 * cls.stat_sum_dataset(dataset)


class CashFitStatistic(FitStatistic):
    """Cash statistic class for Poisson with known background."""

    @classmethod
    def stat_sum_dataset(cls, dataset):
        mask = dataset.mask
        counts, npred = dataset.counts.data, dataset.npred().data

        if mask is not None:
            mask = mask.data.astype("bool")
            counts, npred = counts[mask], npred[mask]

        counts = counts.astype(float)  # This might be done in the Dataset
        return get_fit_statistics_compiled()["cash_sum_compiled"](
            counts.ravel(), npred.ravel()
        )

    @classmethod
    def stat_array_dataset(cls, dataset):
        counts, npred = dataset.counts.data, dataset.npred().data
        return cash(n_on=counts, mu_on=npred)


class WeightedCashFitStatistic(FitStatistic):
    """Cash statistic class for Poisson with known background applying weights."""

    @classmethod
    def stat_sum_dataset(cls, dataset):
        counts, npred = dataset.counts.data.astype(float), dataset.npred().data

        if dataset.mask is not None:
            mask = ~(dataset.mask.data == False)  # noqa
            counts = counts[mask]
            npred = npred[mask]

            weights = dataset.mask.data[mask].astype("float")
            return get_fit_statistics_compiled()["weighted_cash_sum_compiled"](
                counts, npred, weights
            )
        else:
            # No weights back to regular cash statistic
            return get_fit_statistics_compiled()["cash_sum_compiled"](
                counts.ravel(), npred.ravel()
            )

    @classmethod
    def stat_array_dataset(cls, dataset):
        counts, npred = dataset.counts.data, dataset.npred().data
        weights = 1.0
        if dataset.mask is not None:
            weights = dataset.mask.astype("float")
        return cash(n_on=counts, mu_on=npred) * weights


class WStatFitStatistic(FitStatistic):
    """WStat fit statistic class for ON-OFF Poisson measurements."""

    @classmethod
    def stat_array_dataset(cls, dataset):
        """Statistic function value per bin given the current model parameters."""
        counts, counts_off, alpha = (
            dataset.counts.data,
            dataset.counts_off.data,
            dataset.alpha.data,
        )
        npred_signal = dataset.npred_signal().data
        on_stat_ = wstat(
            n_on=counts,
            n_off=counts_off,
            alpha=alpha,
            mu_sig=npred_signal,
        )
        return np.nan_to_num(on_stat_)

    @classmethod
    def stat_sum_dataset(cls, dataset):
        """Statistic function value per bin given the current model parameters."""
        if dataset.counts_off is None and not np.any(dataset.mask_safe.data):
            return 0
        else:
            stat_array = cls.stat_array_dataset(dataset)
            if dataset.mask is not None:
                stat_array = stat_array[dataset.mask.data]
            return np.sum(stat_array)


class Chi2FitStatistic(FitStatistic):
    """Chi2 fit statistic class for measurements with gaussian symmetric errors."""

    @classmethod
    def stat_array_dataset(cls, dataset):
        """Statistic function value per bin given the current model."""
        model = dataset.flux_pred()
        data = dataset.data.dnde.quantity
        try:
            sigma = dataset.data.dnde_err.quantity
        except AttributeError:
            sigma = (dataset.data.dnde_errn + dataset.data.dnde_errp).quantity / 2
        return ((data - model) / sigma).to_value("") ** 2


class Chi2AsymmetricErrorFitStatistic(FitStatistic):
    """Pseudo-Chi2 fit statistic class for measurements with gaussian asymmetric errors with upper limits.

    Assumes that regular data follow asymmetric normal pdf and upper limits follow complementary error functions.
    """

    @classmethod
    def stat_array_dataset(cls, dataset):
        """Estimate statistic from probability distributions,
        assumes that flux points correspond to asymmetric gaussians
        and upper limits complementary error functions.
        """
        model = np.zeros(dataset.data.dnde.data.shape) + dataset.flux_pred().to_value(
            dataset.data.dnde.unit
        )

        stat = np.zeros(model.shape)

        mask_valid = ~np.isnan(dataset.data.dnde.data)
        loc = dataset.data.dnde.data[mask_valid]
        value = model[mask_valid]
        try:
            mask_p = (model >= dataset.data.dnde.data)[mask_valid]
            scale = np.zeros(mask_p.shape)
            scale[mask_p] = dataset.data.dnde_errp.data[mask_valid][mask_p]
            scale[~mask_p] = dataset.data.dnde_errn.data[mask_valid][~mask_p]

            mask_invalid = np.isnan(scale)
            scale[mask_invalid] = dataset.data.dnde_err.data[mask_valid][mask_invalid]
        except AttributeError:
            scale = dataset.data.dnde_err.data[mask_valid]

        stat[mask_valid] = ((value - loc) / scale) ** 2

        mask_ul = dataset.data.is_ul.data
        value = model[mask_ul]
        loc_ul = dataset.data.dnde_ul.data[mask_ul]
        scale_ul = dataset.data.dnde_ul.data[mask_ul]
        stat[mask_ul] = 2 * np.log(
            (erfc((loc_ul - value) / scale_ul) / 2)
            / (erfc((loc_ul - 0) / scale_ul) / 2)
        )

        stat[np.isnan(stat.data)] = 0
        return stat


class ProfileFitStatistic(FitStatistic):
    """Pseudo-Chi2 fit statistic class for measurements with gaussian asymmetric errors with upper limits.

    Assumes that regular data follow asymmetric normal pdf and upper limits follow complementary error functions
    """

    @classmethod
    def stat_array_dataset(cls, dataset):
        """Estimate statistic from interpolation of the likelihood profile."""
        model = np.zeros(dataset.data.dnde.data.shape) + (
            dataset.flux_pred() / dataset.data.dnde_ref
        ).to_value("")
        stat = np.zeros(model.shape)
        for idx in np.ndindex(dataset._profile_interpolators.shape):
            stat[idx] = dataset._profile_interpolators[idx](model[idx])
        return stat



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Model selection."""

from .nested import select_nested_models, NestedModelSelection

__all__ = [
    "select_nested_models",
    "NestedModelSelection",
]



"""
3D map simulation
=================

Simulate a 3D observation of a source with the CTA 1DC response and fit it with the assumed source model.

Prerequisites
-------------

-  Knowledge of 3D extraction and datasets used in gammapy, see for
   example the :doc:`/tutorials/starting/analysis_1` tutorial.

Context
-------

To simulate a specific observation, it is not always necessary to
simulate the full photon list. For many uses cases, simulating directly
a reduced binned dataset is enough: the IRFs reduced in the correct
geometry are combined with a source model to predict an actual number of
counts per bin. The latter is then used to simulate a reduced dataset
using Poisson probability distribution.

This can be done to check the feasibility of a measurement (performance
/ sensitivity study), to test whether fitted parameters really provide a
good fit to the data etc.

Here we will see how to perform a 3D simulation of a CTA observation,
assuming both the spectral and spatial morphology of an observed source.

**Objective: simulate a 3D observation of a source with CTA using the
CTA 1DC response and fit it with the assumed source model.**

Proposed approach
-----------------

Here we can’t use the regular observation objects that are connected to
a `~gammapy.data.DataStore`. Instead, we will create a fake
`~gammapy.data.Observation` that contain some pointing information and
the CTA 1DC IRFs (that are loaded with `~gammapy.irf.load_irf_dict_from_file`).

Next, we will create a `~gammapy.datasets.MapDataset` geometry through
the `~gammapy.makers.MapDatasetMaker`.

Finally, we will define a model consisting of a
`~gammapy.modeling.models.PowerLawSpectralModel` and a
`~gammapy.modeling.models.GaussianSpatialModel`. This model will be assigned to
the dataset and fake the count data.

"""

######################################################################
# Imports and versions
# --------------------
#

import numpy as np
import astropy.units as u
from astropy.coordinates import SkyCoord
import matplotlib.pyplot as plt

# %matplotlib inline
from IPython.display import display
from gammapy.data import FixedPointingInfo, Observation, observatory_locations
from gammapy.datasets import MapDataset
from gammapy.irf import load_irf_dict_from_file
from gammapy.makers import MapDatasetMaker, SafeMaskMaker
from gammapy.maps import MapAxis, WcsGeom
from gammapy.modeling import Fit
from gammapy.modeling.models import (
    FoVBackgroundModel,
    GaussianSpatialModel,
    Models,
    PowerLawSpectralModel,
    SkyModel,
)

######################################################################
# Simulation
# ----------
#


######################################################################
# We will simulate using the CTA-1DC IRFs shipped with gammapy
#

# Loading IRFs
irfs = load_irf_dict_from_file(
    "$GAMMAPY_DATA/cta-1dc/caldb/data/cta/1dc/bcf/South_z20_50h/irf_file.fits"
)

# Define the observation parameters (typically the observation duration and the pointing position):
livetime = 2.0 * u.hr
pointing_position = SkyCoord(0, 0, unit="deg", frame="galactic")
# We want to simulate an observation pointing at a fixed position in the sky.
# For this, we use the `FixedPointingInfo` class
pointing = FixedPointingInfo(
    fixed_icrs=pointing_position.icrs,
)

# Define map geometry for binned simulation
energy_reco = MapAxis.from_edges(
    np.logspace(-1.0, 1.0, 10), unit="TeV", name="energy", interp="log"
)
geom = WcsGeom.create(
    skydir=(0, 0),
    binsz=0.02,
    width=(6, 6),
    frame="galactic",
    axes=[energy_reco],
)
# It is usually useful to have a separate binning for the true energy axis
energy_true = MapAxis.from_edges(
    np.logspace(-1.5, 1.5, 30), unit="TeV", name="energy_true", interp="log"
)

empty = MapDataset.create(geom, name="dataset-simu", energy_axis_true=energy_true)

# Define sky model to used simulate the data.
# Here we use a Gaussian spatial model and a Power Law spectral model.
spatial_model = GaussianSpatialModel(
    lon_0="0.2 deg", lat_0="0.1 deg", sigma="0.3 deg", frame="galactic"
)
spectral_model = PowerLawSpectralModel(
    index=3, amplitude="1e-11 cm-2 s-1 TeV-1", reference="1 TeV"
)
model_simu = SkyModel(
    spatial_model=spatial_model,
    spectral_model=spectral_model,
    name="model-simu",
)

bkg_model = FoVBackgroundModel(dataset_name="dataset-simu")

models = Models([model_simu, bkg_model])
print(models)


######################################################################
# Now, comes the main part of dataset simulation. We create an in-memory
# observation and an empty dataset. We then predict the number of counts
# for the given model, and Poisson fluctuate it using ``fake()`` to make
# a simulated counts maps. Keep in mind that it is important to specify
# the ``selection`` of the maps that you want to produce
#

# Create an in-memory observation
location = observatory_locations["ctao_south"]
obs = Observation.create(
    pointing=pointing, livetime=livetime, irfs=irfs, location=location
)
print(obs)

# Make the MapDataset
maker = MapDatasetMaker(selection=["exposure", "background", "psf", "edisp"])

maker_safe_mask = SafeMaskMaker(methods=["offset-max"], offset_max=4.0 * u.deg)

dataset = maker.run(empty, obs)
dataset = maker_safe_mask.run(dataset, obs)
print(dataset)

# Add the model on the dataset and Poisson fluctuate
dataset.models = models
dataset.fake()
# Do a print on the dataset - there is now a counts maps
print(dataset)


######################################################################
# Now use this dataset as you would in all standard analysis. You can plot
# the maps, or proceed with your custom analysis. In the next section, we
# show the standard 3D fitting as in :doc:`/tutorials/analysis-3d/analysis_3d`
# tutorial.
#

# To plot, eg, counts:
dataset.counts.smooth(0.05 * u.deg).plot_interactive(add_cbar=True, stretch="linear")
plt.show()


######################################################################
# Fit
# ---
#
# In this section, we do a usual 3D fit with the same model used to
# simulated the data and see the stability of the simulations. Often, it
# is useful to simulate many such datasets and look at the distribution of
# the reconstructed parameters.
#

models_fit = models.copy()

# We do not want to fit the background in this case, so we will freeze the parameters
models_fit["dataset-simu-bkg"].spectral_model.norm.frozen = True
models_fit["dataset-simu-bkg"].spectral_model.tilt.frozen = True

dataset.models = models_fit
print(dataset.models)

fit = Fit(optimize_opts={"print_level": 1})
result = fit.run(datasets=[dataset])

dataset.plot_residuals_spatial(method="diff/sqrt(model)", vmin=-0.5, vmax=0.5)
plt.show()


######################################################################
# Compare the injected and fitted models:
#

print(
    "True model: \n",
    model_simu,
    "\n\n Fitted model: \n",
    models_fit["model-simu"],
)


######################################################################
# Get the errors on the fitted parameters from the parameter table
#

display(result.parameters.to_table())



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
import numpy as np
from numpy.testing import assert_allclose
import astropy.units as u
from gammapy.irf import PSFKernel
from gammapy.maps import MapAxis, WcsGeom
from gammapy.modeling.models import DiskSpatialModel
from gammapy.utils.testing import mpl_plot_check


@pytest.fixture
def kernel_gaussian():
    sigma = 0.5 * u.deg
    binsz = 0.1 * u.deg
    geom = WcsGeom.create(
        binsz=binsz, npix=150, axes=[MapAxis((0, 1, 2), unit=u.TeV, name="energy_true")]
    )

    return PSFKernel.from_gauss(geom, sigma)


def test_psf_kernel_from_gauss(kernel_gaussian):
    # Check that both maps are identical
    assert_allclose(
        kernel_gaussian.psf_kernel_map.data[0], kernel_gaussian.psf_kernel_map.data[1]
    )

    # Is there an odd number of pixels
    assert_allclose(np.array(kernel_gaussian.psf_kernel_map.geom.npix) % 2, 1)


def test_psf_kernel_read_write(kernel_gaussian, tmp_path):
    kernel_gaussian.write(tmp_path / "tmp.fits", overwrite=True)
    kernel2 = PSFKernel.read(tmp_path / "tmp.fits")
    assert_allclose(kernel_gaussian.psf_kernel_map.data, kernel2.psf_kernel_map.data)


def test_psf_kernel_to_image():
    sigma1 = 0.5 * u.deg
    sigma2 = 0.2 * u.deg
    binsz = 0.1 * u.deg

    axis = MapAxis.from_energy_bounds(1, 10, 2, unit="TeV", name="energy_true")
    geom = WcsGeom.create(binsz=binsz, npix=50, axes=[axis])

    disk_1 = DiskSpatialModel(r_0=sigma1)
    disk_2 = DiskSpatialModel(r_0=sigma2)

    rad_max = 2.5 * u.deg
    kernel1 = PSFKernel.from_spatial_model(disk_1, geom, max_radius=rad_max, factor=4)
    kernel2 = PSFKernel.from_spatial_model(disk_2, geom, max_radius=rad_max, factor=4)

    kernel1.psf_kernel_map.data[1, :, :] = kernel2.psf_kernel_map.data[1, :, :]

    kernel_image_1 = kernel1.to_image()
    kernel_image_2 = kernel1.to_image(exposure=[1, 2])

    assert_allclose(kernel_image_1.psf_kernel_map.data.sum(), 1.0, atol=1e-5)
    assert_allclose(kernel_image_1.psf_kernel_map.data[0, 25, 25], 0.028096, atol=1e-5)
    assert_allclose(kernel_image_1.psf_kernel_map.data[0, 22, 22], 0.009615, atol=1e-5)
    assert_allclose(kernel_image_1.psf_kernel_map.data[0, 20, 20], 0.0, atol=1e-5)

    assert_allclose(kernel_image_2.psf_kernel_map.data.sum(), 1.0, atol=1e-5)
    assert_allclose(kernel_image_2.psf_kernel_map.data[0, 25, 25], 0.037555, atol=1e-5)
    assert_allclose(kernel_image_2.psf_kernel_map.data[0, 22, 22], 0.007752, atol=1e-5)
    assert_allclose(kernel_image_2.psf_kernel_map.data[0, 20, 20], 0.0, atol=1e-5)


def test_plot_kernel(kernel_gaussian):
    with mpl_plot_check():
        kernel_gaussian.plot_kernel()


def test_peek(kernel_gaussian):
    with mpl_plot_check():
        kernel_gaussian.peek()
